"""
Transportation mode infering
"""
import numpy as np
from changepy import pelt
from changepy.costs import normal_mean

def extract_features(points, n_tops):
    """ Feature extractor

    Args:
        points (:obj:`list` of :obj:`Point`)
        n_tops (int): Number of top speeds to extract
    Returns:
        :obj:`list` of float: with length (n_tops*2). Where the ith even element
            is the ith top speed and the i+1 element is the percentage of time
            spent on that speed
    """
    max_bin = -1
    for point in points:
        max_bin = max(max_bin, point.vel)
    max_bin = int(round(max_bin)) + 1

    # inits histogram
    histogram = [0] * max_bin
    time = 0

    # fills histogram
    for point in points:
        bin_index = int(round(point.vel))
        histogram[bin_index] += point.dt
        time += point.dt

    result = []
    if time == 0:
        return result

    for _ in range(n_tops):
        max_index = np.argmax(histogram)
        value = histogram[max_index] / time
        result.extend([max_index, value])
        histogram[max_index] = -1

    return result

def speed_clustering(clf, points, min_time):
    """ Transportation mode infering, based on changepoint segmentation

    Args:
        clf (:obj:`Classifier`): Classifier to use
        points (:obj:`list` of :obj:`Point`)
        min_time (float): Min time, in seconds, before do another segmentation
    Returns:
        :obj:`list` of :obj:`dict`
    """
    # get changepoint indexes
    data = [p.vel for p in points]
    changepoints = pelt(normal_mean(data, np.std(data)), len(data))

    # Doesn't have change points
    if len(changepoints) == 0:
        changepoints.append(0)

    # insert last point to be a change point
    changepoints.append(len(points) - 1)

    # info for each changepoint
    cp_info = []

    for i in range(0, len(changepoints) - 1):
        from_index = changepoints[i]
        to_index = changepoints[i+1]
        features = extract_features(points[from_index:to_index], clf.feature_length/2)
        if len(features) > 0:
            [probs] = clf.predict([features], verbose=True)
            top_label = sorted(probs.items(), key=lambda val: val[1])
            cp_info.append({
                'from': from_index,
                'to': to_index,
                'classification': probs,
                'label': top_label[-1][0]
                })


    # group based on label
    previous = cp_info[0]
    grouped = []
    # cum_dt = cp_info[0]['dt']

    # TODO: refactor grouping to another function

    for changep in cp_info[1:]:
        if changep['label'] != previous['label']:# and changep['dt'] > min_time:
            previous['to'] = changep['from']
            # previous['dt'] = cum_dt
            grouped.append(previous)
            previous = changep
            # cum_dt = 0
        # cum_dt = cum_dt + changep['dt']
    previous['to'] = cp_info[-1]['to']
    grouped.append(previous)

    return grouped



"""
Util functions
"""
import datetime

PRECISION_PERSON = 5

PRECISION_TABLE = {
    0: [111.32 * 1000, 102.47 * 1000, 78.71 * 1000, 43.496 * 1000],
    1: [11.132 * 1000, 10.247 * 1000, 7.871 * 1000, 4.3496 * 1000],
    2: [1.1132 * 1000, 1.0247 * 1000, 787.1, 434.96 * 1000],
    3: [111.32, 102.47, 78.71, 43.496 * 1000],
    4: [11.132, 10.247, 7.871, 4.3496 * 1000],
    5: [1.1132, 1.0247, 787.1 / 1000, 434.96 / 1000],
    6: [111.32 / 1000, 102.47 / 1000, 78.71 / 1000, 43.496 / 1000],
    7: [11.132 / 1000, 10.247 / 1000, 7.871 / 1000, 4.3496 / 1000],
    8: [1.1132 / 1000, 1.0247 / 1000, 787.1 / (1000 ** 2), 434.96 / (1000 ** 2)]
    }

def estimate_meters_to_deg(meters, precision=PRECISION_PERSON):
    """ Meters to degrees estimation

    See https://en.wikipedia.org/wiki/Decimal_degrees

    Args:
        meters (float)
        precision (float)
    Returns:
        float: meters in degrees approximation
    """
    line = PRECISION_TABLE[precision]
    dec = 1/float(10 ** precision)
    return meters / line[3] * dec

def isostr_to_datetime(dt_str):
    """ Converts iso formated text string into a datetime object

    Args:
        dt_str (str): ISO formated text string
    Returns:
        :obj:`datetime.datetime`
    """
    delta_t, _, us_m = dt_str.partition(".")
    delta_t = datetime.datetime.strptime(delta_t, "%Y-%m-%dT%H:%M:%S")
    us_m = int(us_m.rstrip("Z"), 10)
    return delta_t + datetime.timedelta(microseconds=us_m)



import sys
import contextlib
from distutils.errors import DistutilsOptionError
from unittest import TestLoader

from setuptools.extern import six
from setuptools.extern.six.moves import map

from pkg_resources import (resource_listdir, resource_exists, normalize_path,
                           working_set, _namespace_packages,
                           add_activation_listener, require, EntryPoint)
from setuptools import Command
from setuptools.py31compat import unittest_main


class ScanningLoader(TestLoader):

    def loadTestsFromModule(self, module, pattern=None):
        """Return a suite of all tests cases contained in the given module

        If the module is a package, load tests from all the modules in it.
        If the module has an ``additional_tests`` function, call it and add
        the return value to the tests.
        """
        tests = []
        tests.append(TestLoader.loadTestsFromModule(self, module))

        if hasattr(module, "additional_tests"):
            tests.append(module.additional_tests())

        if hasattr(module, '__path__'):
            for file in resource_listdir(module.__name__, ''):
                if file.endswith('.py') and file != '__init__.py':
                    submodule = module.__name__ + '.' + file[:-3]
                else:
                    if resource_exists(module.__name__, file + '/__init__.py'):
                        submodule = module.__name__ + '.' + file
                    else:
                        continue
                tests.append(self.loadTestsFromName(submodule))

        if len(tests) != 1:
            return self.suiteClass(tests)
        else:
            return tests[0]  # don't create a nested suite for only one return


# adapted from jaraco.classes.properties:NonDataProperty
class NonDataProperty(object):

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return self.fget(obj)


class test(Command):
    """Command to run unit tests after in-place build"""

    description = "run unit tests after in-place build"

    user_options = [
        ('test-module=', 'm', "Run 'test_suite' in specified module"),
        ('test-suite=', 's',
         "Test suite to run (e.g. 'some_module.test_suite')"),
        ('test-runner=', 'r', "Test runner to use"),
    ]

    def initialize_options(self):
        self.test_suite = None
        self.test_module = None
        self.test_loader = None
        self.test_runner = None

    def finalize_options(self):

        if self.test_suite and self.test_module:
            msg = "You may specify a module or a suite, but not both"
            raise DistutilsOptionError(msg)

        if self.test_suite is None:
            if self.test_module is None:
                self.test_suite = self.distribution.test_suite
            else:
                self.test_suite = self.test_module + ".test_suite"

        if self.test_loader is None:
            self.test_loader = getattr(self.distribution, 'test_loader', None)
        if self.test_loader is None:
            self.test_loader = "setuptools.command.test:ScanningLoader"
        if self.test_runner is None:
            self.test_runner = getattr(self.distribution, 'test_runner', None)

    @NonDataProperty
    def test_args(self):
        return list(self._test_args())

    def _test_args(self):
        if self.verbose:
            yield '--verbose'
        if self.test_suite:
            yield self.test_suite

    def with_project_on_sys_path(self, func):
        """
        Backward compatibility for project_on_sys_path context.
        """
        with self.project_on_sys_path():
            func()

    @contextlib.contextmanager
    def project_on_sys_path(self):
        with_2to3 = six.PY3 and getattr(self.distribution, 'use_2to3', False)

        if with_2to3:
            # If we run 2to3 we can not do this inplace:

            # Ensure metadata is up-to-date
            self.reinitialize_command('build_py', inplace=0)
            self.run_command('build_py')
            bpy_cmd = self.get_finalized_command("build_py")
            build_path = normalize_path(bpy_cmd.build_lib)

            # Build extensions
            self.reinitialize_command('egg_info', egg_base=build_path)
            self.run_command('egg_info')

            self.reinitialize_command('build_ext', inplace=0)
            self.run_command('build_ext')
        else:
            # Without 2to3 inplace works fine:
            self.run_command('egg_info')

            # Build extensions in-place
            self.reinitialize_command('build_ext', inplace=1)
            self.run_command('build_ext')

        ei_cmd = self.get_finalized_command("egg_info")

        old_path = sys.path[:]
        old_modules = sys.modules.copy()

        try:
            sys.path.insert(0, normalize_path(ei_cmd.egg_base))
            working_set.__init__()
            add_activation_listener(lambda dist: dist.activate())
            require('%s==%s' % (ei_cmd.egg_name, ei_cmd.egg_version))
            yield
        finally:
            sys.path[:] = old_path
            sys.modules.clear()
            sys.modules.update(old_modules)
            working_set.__init__()

    def run(self):
        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(
                self.distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)

        cmd = ' '.join(self._argv)
        if self.dry_run:
            self.announce('skipping "%s" (dry run)' % cmd)
            return

        self.announce('running "%s"' % cmd)
        with self.project_on_sys_path():
            self.run_tests()

    def run_tests(self):
        # Purge modules under test from sys.modules. The test loader will
        # re-import them from the build location. Required when 2to3 is used
        # with namespace packages.
        if six.PY3 and getattr(self.distribution, 'use_2to3', False):
            module = self.test_suite.split('.')[0]
            if module in _namespace_packages:
                del_modules = []
                if module in sys.modules:
                    del_modules.append(module)
                module += '.'
                for name in sys.modules:
                    if name.startswith(module):
                        del_modules.append(name)
                list(map(sys.modules.__delitem__, del_modules))

        unittest_main(
            None, None, self._argv,
            testLoader=self._resolve_as_ep(self.test_loader),
            testRunner=self._resolve_as_ep(self.test_runner),
        )

    @property
    def _argv(self):
        return ['unittest'] + self.test_args

    @staticmethod
    def _resolve_as_ep(val):
        """
        Load the indicated attribute value, called, as a as if it were
        specified as an entry point.
        """
        if val is None:
            return
        parsed = EntryPoint.parse("x=" + val)
        return parsed.resolve()()



"""
Douglas ramer peucker
"""

from math import sqrt

def distance(p_a, p_b):
    """ Euclidean distance, between two points

    Args:
        p_a (:obj:`Point`)
        p_b (:obj:`Point`)
    Returns:
        float: distance, in degrees
    """
    return sqrt((p_a.lat - p_b.lat) ** 2 + (p_a.lon - p_b.lon) ** 2)

def point_line_distance(point, start, end):
    """ Distance from a point to a line, formed by two points

    Args:
        point (:obj:`Point`)
        start (:obj:`Point`): line point
        end (:obj:`Point`): line point
    Returns:
        float: distance to line, in degrees
    """
    if start == end:
        return distance(point, start)
    else:
        un_dist = abs(
            (end.lat-start.lat)*(start.lon-point.lon) - (start.lat-point.lat)*(end.lon-start.lon)
        )
        n_dist = sqrt(
            (end.lat-start.lat)**2 + (end.lon-start.lon)**2
        )
        return un_dist / n_dist

def drp(points, epsilon):
    """ Douglas ramer peucker

    Based on https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm

    Args:
        points (:obj:`list` of :obj:`Point`)
        epsilon (float): drp threshold
    Returns:
        :obj:`list` of :obj:`Point`
    """
    dmax = 0.0
    index = 0

    for i in range(1, len(points)-1):
        dist = point_line_distance(points[i], points[0], points[-1])
        if dist > dmax:
            index = i
            dmax = dist

    if dmax > epsilon:
        return drp(points[:index+1], epsilon)[:-1] + drp(points[index:], epsilon)
    else:
        return [points[0], points[-1]]



import sys
from os import listdir
from os.path import join
import numpy as np
from tracktotrip import Track
from tracktotrip.classifier import Classifier
from tracktotrip.transportation_mode import extract_features
import pickle

def process_track(filename):
    t = Track.from_gpx(filename)[0]
    t.compute_metrics()

    for segment in t.segments:
        features = extract_features(segment.points, 4)
    return features

def process_folder(folder, alias={}, skip=[]):
    files = listdir(folder)
    files = [f for f in files if f.endswith('.gpx')]

    labels = []
    features = []
    alias_keys = list(alias.keys())
    l = len(files)
    for i, gpx in enumerate(files):
        sys.stdout.write('Processing %d of %d: %s\r' % (i + 1, l, gpx))
        sys.stdout.flush()

        label = gpx.split('.')[0]
        label = label.lower()
        if label in skip:
            continue

        if label in alias_keys:
            label = alias[label]

        feature = process_track(join(folder, gpx))

        if len(feature) > 0:
            labels.append(label)
            features.append(feature)

    return features, labels

features, labels = process_folder('./dataset', alias={
    'bus': 'vehicle',
    'car': 'vehicle',
    'taxi:': 'vehicle'
}, skip=['bike'])

print('Saving features and labels')
pickle.dump(features, open('geolife.3features', 'w'))
pickle.dump(labels, open('geolife.3labels', 'w'))

print('Classifier built')
clf = Classifier()
clf.learn(features, labels)




"""
Point class
"""
import math
import datetime
from .utils import isostr_to_datetime

EPOCH = datetime.datetime.utcfromtimestamp(0)

class Point(object):
    """ Spaciotemporal point representation

    Attributes:
        lat (float): latitude
        lon (float): longitude
        time (:obj:`datetime.datetime`): time
        dt (float): time difference in seconds from the past point in the segment
            should be computed
        acc (float): accelaration in km^2/h, relative to the previous point in the segment
            should be computed with compute_metrics method
        vel (float): velocity in km/h, relative to the previous point in the segment
            should be computed with compute_metrics method
    """
    def __init__(self, lat, lon, time):
        self.lon = lon
        self.lat = lat
        self.time = time
        self.dt = .0 #pylint: disable=invalid-name
        self.acc = .0
        self.vel = .0

    def get_timestamp(self):
        """ Gets the timestamp of this point's time, seconds since 1970

        Returns:
            float: time since epoch, in seconds
        """
        return (self.time - EPOCH).total_seconds()

    def gen2arr(self):
        """ Generate a location array

        Returns:
            :obj:`list` of float: List with longitude and latitude
        """
        return [self.lon, self.lat]

    def gen3arr(self):
        """ Generate a time-location array

        Returns:
            :obj:`list` of float: List with longitude, latitude and timestamp
        """
        return [self.lon, self.lat, self.get_timestamp()]

    def distance(self, other):
        """ Distance between points

        Args:
            other (:obj:`Point`)
        Returns:
            float: Distance in km
        """
        return distance(self.lat, self.lon, None, other.lat, other.lon, None)

    def time_difference(self, previous):
        """ Calcultes the time difference against another point

        Args:
            previous (:obj:`Point`): Point before
        Returns:
            Time difference in seconds
        """
        return abs((self.time - previous.time).total_seconds())

    def compute_metrics(self, previous):
        """ Computes the metrics of this point

        Computes and updates the dt, vel and acc attributes.

        Args:
            previous (:obj:`Point`): Point before
        Returns:
            :obj:`Point`: Self
        """
        delta_t = self.time_difference(previous)
        vel = 0
        delta_v = 0
        acc = 0
        if delta_t != 0:
            vel = self.distance(previous)/delta_t
            delta_v = vel - previous.vel
            acc = delta_v/delta_t

        self.dt = delta_t
        self.acc = acc
        self.vel = vel
        return self

    @staticmethod
    def from_gpx(gpx_track_point):
        """ Creates a point from GPX representation

        Arguments:
            gpx_track_point (:obj:`gpxpy.GPXTrackPoint`)
        Returns:
            :obj:`Point`
        """
        return Point(
            lat=gpx_track_point.latitude,
            lon=gpx_track_point.longitude,
            time=gpx_track_point.time
        )

    def to_json(self):
        """ Creates a JSON serializable representation of this instance

        Returns:
            :obj:`dict`: For example,
                {
                    "lat": 9.3470298,
                    "lon": 3.79274,
                    "time": "2016-07-15T15:27:53.574110"
                }
        """
        return {
            'lat': self.lat,
            'lon': self.lon,
            'time': self.time.isoformat()
        }

    @staticmethod
    def from_json(json):
        """ Creates Point instance from JSON representation

        Args:
            json (:obj:`dict`): Must have at least the following keys: lat (float), lon (float),
                time (string in iso format). Example,
                {
                    "lat": 9.3470298,
                    "lon": 3.79274,
                    "time": "2016-07-15T15:27:53.574110"
                }
            json: map representation of Point instance
        Returns:
            :obj:`Point`
        """
        return Point(
            lat=json['lat'],
            lon=json['lon'],
            time=isostr_to_datetime(json['time'])
        )


ONE_DEGREE = 1000. * 10000.8 / 90.
EARTH_RADIUS = 6371 * 1000

def to_rad(number):
    """ Degrees to rads """
    return number / 180. * math.pi

def haversine_distance(latitude_1, longitude_1, latitude_2, longitude_2):
    """
    Haversine distance between two points, expressed in meters.
    Implemented from http://www.movable-type.co.uk/scripts/latlong.html
    """
    d_lat = to_rad(latitude_1 - latitude_2)
    d_lon = to_rad(longitude_1 - longitude_2)
    lat1 = to_rad(latitude_1)
    lat2 = to_rad(latitude_2)

    #pylint: disable=invalid-name
    a = math.sin(d_lat/2) * math.sin(d_lat/2) + \
        math.sin(d_lon/2) * math.sin(d_lon/2) * math.cos(lat1) * math.cos(lat2)
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1-a))
    d = EARTH_RADIUS * c

    return d

#pylint: disable=too-many-arguments
def distance(latitude_1, longitude_1, elevation_1, latitude_2, longitude_2, elevation_2,
             haversine=None):
    """ Distance between two points """

    # If points too distant -- compute haversine distance:
    if haversine or (abs(latitude_1 - latitude_2) > .2 or abs(longitude_1 - longitude_2) > .2):
        return haversine_distance(latitude_1, longitude_1, latitude_2, longitude_2)

    coef = math.cos(latitude_1 / 180. * math.pi)
    #pylint: disable=invalid-name
    x = latitude_1 - latitude_2
    y = (longitude_1 - longitude_2) * coef

    distance_2d = math.sqrt(x * x + y * y) * ONE_DEGREE

    if elevation_1 is None or elevation_2 is None or elevation_1 == elevation_2:
        return distance_2d

    return math.sqrt(distance_2d ** 2 + (elevation_1 - elevation_2) ** 2)



"""
Transportation mode infering
"""
import numpy as np
from changepy import pelt
from changepy.costs import normal_mean
from .utils import pairwise

def learn_transportation_mode(track, clf):
    """ Inserts transportation modes of a track into a classifier

    Args:
        track (:obj:`Track`)
        clf (:obj:`Classifier`)
    """
    for segment in track.segments:
        tmodes = segment.transportation_modes
        points = segment.points
        features = []
        labels = []

        for tmode in tmodes:
            points_part = points[tmode['from']:tmode['to']]
            if len(points_part) > 0:
                features.append(extract_features(points_part, clf.feature_length/2))
                labels.append(tmode['label'])

        clf.learn(features, labels)

def extract_features(points, n_tops):
    """ Feature extractor

    Args:
        points (:obj:`list` of :obj:`Point`)
        n_tops (int): Number of top speeds to extract
    Returns:
        :obj:`list` of float: with length (n_tops*2). Where the ith even element
            is the ith top speed and the i+1 element is the percentage of time
            spent on that speed
    """
    max_bin = -1
    for point in points:
        max_bin = max(max_bin, point.vel)
    max_bin = int(round(max_bin)) + 1

    # inits histogram
    histogram = [0] * max_bin
    time = 0

    # fills histogram
    for point in points:
        bin_index = int(round(point.vel))
        histogram[bin_index] += point.dt
        time += point.dt

    result = []
    if time == 0:
        return result

    for _ in range(n_tops):
        max_index = np.argmax(histogram)
        value = histogram[max_index] / time
        result.extend([max_index, value])
        histogram[max_index] = -1

    return result

def speed_difference(points):
    """ Computes the speed difference between each adjacent point

    Args:
        points (:obj:`Point`)
    Returns:
        :obj:`list` of int: Indexes of changepoints
    """
    data = [0]
    for before, after in pairwise(points):
        data.append(before.vel - after.vel)
    return data

def detect_changepoints(points, min_time, data_processor=speed_difference):
    """ Detects changepoints on points that have at least a specific duration

    Args:
        points (:obj:`Point`)
        min_time (float): Min time that a sub-segmented, bounded by two changepoints, must have
        data_processor (function): Function to extract data to feed to the changepoint algorithm.
            Defaults to `speed_difference`
    Returns:
        :obj:`list` of int: Indexes of changepoints
    """
    data = data_processor(points)
    changepoints = pelt(normal_mean(data, np.std(data)), len(data))
    changepoints.append(len(points) - 1)

    result = []
    for start, end in pairwise(changepoints):
        time_diff = points[end].time_difference(points[start])
        if time_diff > min_time:
            result.append(start)

    # adds the last changepoint detected
    result.append(changepoints[-1])
    return sorted(list(set(result)))

def group_modes(modes):
    """ Groups consecutive transportation modes with same label, into one

    Args:
        modes (:obj:`list` of :obj:`dict`)
    Returns:
        :obj:`list` of :obj:`dict`
    """
    previous = modes[0]
    grouped = []

    for changep in modes[1:]:
        if changep['label'] != previous['label']:
            previous['to'] = changep['from']
            grouped.append(previous)
            previous = changep

    previous['to'] = modes[-1]['to']
    grouped.append(previous)
    return grouped

def speed_clustering(clf, points, min_time):
    """ Transportation mode infering, based on changepoint segmentation

    Args:
        clf (:obj:`Classifier`): Classifier to use
        points (:obj:`list` of :obj:`Point`)
        min_time (float): Min time, in seconds, before do another segmentation
    Returns:
        :obj:`list` of :obj:`dict`
    """
    # get changepoint indexes
    changepoints = detect_changepoints(points, min_time)

    # info for each changepoint
    cp_info = []

    for i in range(0, len(changepoints) - 1):
        from_index = changepoints[i]
        to_index = changepoints[i+1]
        features = extract_features(points[from_index:to_index], clf.feature_length/2)
        if len(features) > 0:
            [probs] = clf.predict([features], verbose=True)
            top_label = sorted(probs.items(), key=lambda val: val[1])
            cp_info.append({
                'from': from_index,
                'to': to_index,
                'classification': probs,
                'label': top_label[-1][0]
                })

    return group_modes(cp_info)



"""
Similarity functions
"""
import math
import numpy as np
from rtree import index

#pylint: disable=invalid-name

def dot(p1, p2):
    """Dot product between two points

    Args:
        p1 ([float, float]): x and y coordinates
        p2 ([float, float]): x and y coordinates
    Returns:
        float
    """
    return p1[0] * p2[0] + p1[1] * p2[1]

def normalize(p):
    """Normalizes a point/vector

    Args:
        p ([float, float]): x and y coordinates
    Returns:
        float
    """
    l = math.sqrt(p[0]**2 + p[1]**2)
    return [p[0]/l, p[1]/l]

def angle(p1, p2):
    """Angle between two points

    Args:
        p1 ([float, float]): x and y coordinates
        p2 ([float, float]): x and y coordinates
    Returns:
        float
    """
    return dot(p1, p2)

def angle_similarity(l1, l2):
    """Computes the similarity between two lines, based
    on their angles

    Args:
        l1 ([float, float]): x and y coordinates
        l2 ([float, float]): x and y coordinates
    Returns:
        float
    """
    return angle(l1, l2)

def line(p1, p2):
    """Creates a line from two points

    From http://stackoverflow.com/a/20679579

    Args:
        p1 ([float, float]): x and y coordinates
        p2 ([float, float]): x and y coordinates
    Returns:
        (float, float, float): x, y and _
    """
    A = (p1[1] - p2[1])
    B = (p2[0] - p1[0])
    C = (p1[0]*p2[1] - p2[0]*p1[1])
    return A, B, -C

def intersection(L1, L2):
    """Intersects two line segments

    Args:
        L1 ([float, float]): x and y coordinates
        L2 ([float, float]): x and y coordinates
    Returns:
        bool: if they intersect
        (float, float): x and y of intersection, if they do
    """
    D = L1[0] * L2[1] - L1[1] * L2[0]
    Dx = L1[2] * L2[1] - L1[1] * L2[2]
    Dy = L1[0] * L2[2] - L1[2] * L2[0]
    if D != 0:
        x = Dx / D
        y = Dy / D
        return x, y
    else:
        return False

def distance(a, b):
    """Distance between two points

    Args:
        a ([float, float]): x and y coordinates
        b ([float, float]): x and y coordinates
    Returns:
        float
    """
    return math.sqrt((b[0]-a[0])**2 + (b[1]-a[1])**2)

def distance_tt_point(a, b):
    """ Euclidean distance between two (tracktotrip) points

    Args:
        a (:obj:`Point`)
        b (:obj:`Point`)
    Returns:
        float
    """
    return math.sqrt((b.lat-a.lat)**2 + (b.lon-a.lon)**2)

def closest_point(a, b, p):
    """Finds closest point in a line segment

    Args:
        a ([float, float]): x and y coordinates. Line start
        b ([float, float]): x and y coordinates. Line end
        p ([float, float]): x and y coordinates. Point to find in the segment
    Returns:
        (float, float): x and y coordinates of the closest point
    """
    ap = [p[0]-a[0], p[1]-a[1]]
    ab = [b[0]-a[0], b[1]-a[1]]
    mag = float(ab[0]**2 + ab[1]**2)
    proj = dot(ap, ab)
    dist = proj / mag
    if dist < 0:
        return [a[0], a[1]]
    elif dist > 1:
        return [b[0], b[1]]
    else:
        return [a[0] + ab[0] * dist, a[1] + ab[1] * dist]

def distance_to_line(a, b, p):
    """Closest distance between a line segment and a point

    Args:
        a ([float, float]): x and y coordinates. Line start
        b ([float, float]): x and y coordinates. Line end
        p ([float, float]): x and y coordinates. Point to compute the distance
    Returns:
        float
    """
    return distance(closest_point(a, b, p), p)

CLOSE_DISTANCE_THRESHOLD = 1.0
def distance_similarity(a, b, p, T=CLOSE_DISTANCE_THRESHOLD):
    """Computes the distance similarity between a line segment
    and a point

    Args:
        a ([float, float]): x and y coordinates. Line start
        b ([float, float]): x and y coordinates. Line end
        p ([float, float]): x and y coordinates. Point to compute the distance
    Returns:
        float: between 0 and 1. Where 1 is very similar and 0 is completely different
    """
    d = distance_to_line(a, b, p)
    r = (-1/float(T)) * abs(d) + 1

    return r if r > 0 else 0

def line_distance_similarity(p1a, p1b, p2a, p2b, T=CLOSE_DISTANCE_THRESHOLD):
    """Line distance similarity between two line segments

    Args:
        p1a ([float, float]): x and y coordinates. Line A start
        p1b ([float, float]): x and y coordinates. Line A end
        p2a ([float, float]): x and y coordinates. Line B start
        p2b ([float, float]): x and y coordinates. Line B end
    Returns:
        float: between 0 and 1. Where 1 is very similar and 0 is completely different
    """
    d1 = distance_similarity(p1a, p1b, p2a, T=T)
    d2 = distance_similarity(p1a, p1b, p2b, T=T)
    return abs(d1 + d2) * 0.5

def line_similarity(p1a, p1b, p2a, p2b, T=CLOSE_DISTANCE_THRESHOLD):
    """Similarity between two lines

    Args:
        p1a ([float, float]): x and y coordinates. Line A start
        p1b ([float, float]): x and y coordinates. Line A end
        p2a ([float, float]): x and y coordinates. Line B start
        p2b ([float, float]): x and y coordinates. Line B end
    Returns:
        float: between 0 and 1. Where 1 is very similar and 0 is completely different
    """
    d = line_distance_similarity(p1a, p1b, p2a, p2b, T=T)
    a = abs(angle_similarity(normalize(line(p1a, p1b)), normalize(line(p2a, p2b))))
    return d * a

def bounding_box_from(points, i, i1):
    """Creates bounding box for a line segment

    Args:
        points (:obj:`list` of :obj:`Point`)
        i (int): Line segment start, index in points array
        i1 (int): Line segment end, index in points array
    Returns:
        (float, float, float, float): with bounding box min x, min y, max x and max y
    """
    pi = points[i]
    pi1 = points[i1]

    min_lat = min(pi.lat, pi1.lat)
    min_lon = min(pi.lon, pi1.lon)
    max_lat = max(pi.lat, pi1.lat)
    max_lon = max(pi.lon, pi1.lon)

    latd = (max_lat-min_lat) * 2
    lond = (max_lon-min_lon) * 2

    return min_lat-latd, min_lon-lond, max_lat+latd, max_lon+lond

def segment_similarity(A, B):
    """Computes the similarity between two segments

    Args:
        A (:obj:`Segment`)
        B (:obj:`Segment`)
    Returns:
        float: between 0 and 1. Where 1 is very similar and 0 is completely different
    """
    l_a = len(A.points)
    l_b = len(B.points)

    idx = index.Index()
    dex = 0
    for i in range(l_a-1):
        idx.insert(dex, bounding_box_from(A.points, i, i+1), obj=[A.points[i], A.points[i+1]])
        dex = dex + 1

    prox_acc = []

    for i in range(l_b-1):
        ti = B.points[i].gen2arr()
        ti1 = B.points[i+1].gen2arr()
        bb = bounding_box_from(B.points, i, i+1)
        intersects = idx.intersection(bb, objects=True)
        n_prox = []
        i_prox = 0
        a = 0
        for x in intersects:
            a = a + 1
            pi = x.object[0].gen2arr()
            pi1 = x.object[1].gen2arr()
            prox = line_similarity(ti, ti1, pi, pi1)
            i_prox = i_prox + prox
            n_prox.append(prox)

        if a != 0:
            # prox_acc.append(i_prox / a)
            prox_acc.append(max(n_prox))
        else:
            prox_acc.append(0)

    return np.mean(prox_acc), prox_acc

def sort_segment_points(Aps, Bps):
    """Takes two line segments and sorts all their points,
    so that they form a continuous path

    Args:
        Aps: Array of tracktotrip.Point
        Bps: Array of tracktotrip.Point
    Returns:
        Array with points ordered
    """
    mid = []
    j = 0
    mid.append(Aps[0])
    for i in range(len(Aps)-1):
        dist = distance_tt_point(Aps[i], Aps[i+1])
        for m in range(j, len(Bps)):
            distm = distance_tt_point(Aps[i], Bps[m])
            if dist < distm:
                j = m
                break
            mid.append(Bps[m])

        mid.append(Aps[i+1])
    for m in range(j, len(Bps)):
        mid.append(Bps[m])
    return mid



import sys
import gpxpy
import gpxpy.gpx
import datetime
import argparse
import urllib2, zipfile, StringIO
from os.path import join
from os import listdir

def ctime(day, hour="", sep="-"):
    t = day
    if len(hour) != 0:
        t += " " + hour
    return datetime.datetime.strptime(t, "%Y"+sep+"%m"+sep+"%d %H:%M:%S")

def readLabels(folder):
    labels = {}
    with open(join(folder, 'labels.txt'), 'r') as f:
        for line in f.readlines()[1:]:
            start, end, label = line.strip().split('\t')
            if not(label in labels):
                labels[label] = []
            start = ctime(start, sep="/")
            end = ctime(end, sep="/")
            labels[label].append((start, end))
    return labels

def readPlt(name):
    points = []
    with open(name, 'r') as f:
        for line in f.readlines()[6:]:
            lat, lon, _, alt, _, day, hour = line.strip().split(',')
            points.append({'lat': float(lat), 'lon': float(lon), 'alt': float(alt), 'time': ctime(day, hour)})
    return points

def extractSegmentsFromFile(f, labels):
    points = readPlt(f)
    _segments = {}
    table = []
    for label, segments in labels.iteritems():
        for segment in segments:
            table.append((label, segment[0], segment[1]))

    for point in points:
        ptime = point['time']
        for i, segment in enumerate(table):
            label, start, end = segment
            if start <= ptime and ptime <= end:
                if not(i in _segments):
                    _segments[i] = (label, [])
                _segments[i][1].append(point)
    return _segments.values()

def saveFile(folder, filename, content):
    name = join(folder, filename)
    with open(name, 'w') as f:
        f.write(content)
        f.close()

def saveSegments(segments, f, folder):
    name = f.split('/')[-1]
    for i, segment in enumerate(segments):
        label, points = segment
        filename = label + '.' + str(i) + '.' + str(len(points)) + '.' + name.split('.')[0] + '.gpx'

        gpx = gpxpy.gpx.GPX()

        gpx_track = gpxpy.gpx.GPXTrack()
        gpx.tracks.append(gpx_track)

        gpx_segment = gpxpy.gpx.GPXTrackSegment()
        gpx_track.segments.append(gpx_segment)

        for point in points:
            gpx_segment.points.append(gpxpy.gpx.GPXTrackPoint(latitude=point['lat'], longitude=point['lon'], elevation=point['alt'], time=point['time']))

        saveFile(folder, filename, gpx.to_xml())

def processFolderGroup(folder, output):
    # format: { walk: [ (start-time, end-time), ... ], ... }
    try:
        labels = readLabels(folder)

        trajFolder = join(folder, 'Trajectory')
        fileList = filter(lambda f: f.split('.')[-1] == 'plt', listdir(trajFolder))

        l = len(fileList)
        for i, f in enumerate(fileList):

            sys.stdout.write('Processing %d of %d: %s\r' % (i + 1, l, f))
            sys.stdout.flush()
            # format: [( label, segment ), ...]
            #       segment is : [ (lat, lon, time), ... ]
            segments = extractSegmentsFromFile(join(trajFolder, f), labels)
            if len(segments) > 0:
                saveSegments(segments, f, output)

        sys.stdout.write('Processed\n')
        sys.stdout.flush()
    except IOError:
        print('Skipping')
        return
    except KeyboardInterrupt:
        exit()
    except:
        print(sys.exc_info()[0])

def processDataset(root = 'Data', output = 'out'):
    folders = listdir(root)
    for folder in folders:
        f = join(root, folder)
        print('Going to ' + f)
        processFolderGroup(f, output)
    print('Done')

lastTime = None
def chunk_report(bytes_so_far, chunk_size, total_size):
   percent = float(bytes_so_far) / total_size
   percent = round(percent*100, 2)
   mb = 1024 * 1024
   sys.stdout.write("Downloaded %d of %dMB (%0.2f%%)\r" %
       (bytes_so_far / mb, total_size / mb, percent))

   if bytes_so_far >= total_size:
      sys.stdout.write('\n')

def chunk_read(response, chunk_size=8192, report_hook=None):
   total_size = response.info().getheader('Content-Length').strip()
   total_size = int(total_size)
   bytes_so_far = 0

   while 1:
      chunk = response.read(chunk_size)
      bytes_so_far += len(chunk)

      if not chunk:
         break

      if report_hook:
         report_hook(bytes_so_far, chunk_size, total_size)

   return bytes_so_far

def downloadDataset(toFolder):
    print('Dowloading GeoLife dataset from http://research.microsoft.com/en-us/downloads/b16d359d-d164-469e-9fd4-daa38f2b2e13/')
    print('This might take a few minutes, the file size is approximately 300MB')
    r = urllib2.urlopen('http://ftp.research.microsoft.com/downloads/b16d359d-d164-469e-9fd4-daa38f2b2e13/Geolife%20Trajectories%201.3.zip')
    chunk_read(r, report_hook=chunk_report)
    z = zipfile.ZipFile(StringIO.StringIO(r))
    z.extractall(toFolder)

parser = argparse.ArgumentParser(description='GeoLife Trajectory dataset transportation mode extractor.\n\nExtracts transportation mode from the dataset, into individual files,\nannotated with the following format:\n\t[transporation mode].[control].[nPoints].[original-file].gpx')
parser.add_argument('datasetFolder', metavar='datasetFolder', type=str, help='Path to the GeoLife dataset folder')
parser.add_argument('-o', '--output', metavar='outputFolder', type=str, help='Path to processed dataset')
parser.add_argument('-d', '--download', dest='download', action='store_true', required=False, help='Pass this flag to download the GeoLife dataset to the specified folder and to process it')
args = parser.parse_args()

print(args)
if args.download == True:
    downloadDataset(args.datasetFolder)
else:
    processDataset(args.datasetFolder, args.output)



from tracktotrip import learn_trip, Track, Segment, Point
from datetime import datetime, timedelta
import unittest

class TestLearnTrip(unittest.TestCase):
    def setUp(self):
        Aps = [[0.5, 0.5], [1, 1.5], [2, 2.5], [3.5, 3.5], [5.2, 4.5], [7.5, 6.5], [7.9, 8]]
        Bps = [[0.6, 0.5], [1.05, 1.45], [2.1, 2.4], [2.8, 4], [3.5, 5.5], [5, 5.7], [7.8, 5.7], [8.1, 6.5], [8.1, 8]]

        time = datetime.now()
        dt = timedelta(1000)

        def pt_arr_to_track(pts):
            seg = Segment(map(lambda p: Point(None, p[0], p[1], time + dt), pts))
            return Track(name="TripA", segments=[seg])

        self.tripA = pt_arr_to_track(Aps)
        self.tripA.toTrip(name="A")

        self.tripB = pt_arr_to_track(Bps)
        self.tripB.toTrip(name="B")

    def fail(self, track):
        self.assertTrue(False)


    def test_empty_list(self):
        # print(len(self.tripA.segments[0].points))
        def expected(track):
            # print(len(track.segments[0].points))
            self.assertTrue(True)

        learn_trip.learn_trip(self.tripA, [], expected, self.fail)

    def test_exact_matching_track(self):
        def expected(track):
            self.assertTrue(True)

        learn_trip.learn_trip(self.tripA, [self.tripA.copy()], self.fail, expected)

    def test_matching_track(self):
        def expected(track):
            self.assertTrue(True)

        learn_trip.learn_trip(self.tripA, [self.tripB], self.fail, expected)


if __name__ == '__main__':
    unittest.main()



"""
This module adds improved support for Microsoft Visual C++ compilers.
"""
import os
import platform
import itertools
import distutils.errors

from setuptools.extern.six.moves import filterfalse

if platform.system() == 'Windows':
    from setuptools.extern.six.moves import winreg
    safe_env = os.environ
else:
    """
    Mock winreg and environ so the module can be imported
    on this platform.
    """
    class winreg:
        HKEY_USERS = None
        HKEY_CURRENT_USER = None
        HKEY_LOCAL_MACHINE = None
        HKEY_CLASSES_ROOT = None
    safe_env = dict()

try:
    # Distutil file for MSVC++ 9.0 and upper (Python 2.7 to 3.4)
    import distutils.msvc9compiler as msvc9compiler
except ImportError:
    pass

try:
    # Distutil file for MSVC++ 14.0 and upper (Python 3.5+)
    import distutils._msvccompiler as msvc14compiler
except ImportError:
    pass


unpatched = dict()


def patch_for_specialized_compiler():
    """
    Patch functions in distutils to use standalone Microsoft Visual C++
    compilers.

    Known supported compilers:
    --------------------------
    Microsoft Visual C++ 9.0:
        Microsoft Visual C++ Compiler for Python 2.7 (x86, amd64);
        Microsoft Windows SDK 7.0 (x86, x64, ia64);
        Microsoft Windows SDK 6.1 (x86, x64, ia64)

    Microsoft Visual C++ 10.0:
        Microsoft Windows SDK 7.1 (x86, x64, ia64)

    Microsoft Visual C++ 14.0:
        Microsoft Visual C++ Build Tools 2015 (x86, x64, arm)
    """
    if platform.system() != 'Windows':
        # Compilers only availables on Microsoft Windows
        return

    if 'distutils' not in globals():
        # The module isn't available to be patched
        return

    if unpatched:
        # Already patched
        return

    try:
        # Patch distutils.msvc9compiler
        unpatched['msvc9_find_vcvarsall'] = msvc9compiler.find_vcvarsall
        msvc9compiler.find_vcvarsall = msvc9_find_vcvarsall
        unpatched['msvc9_query_vcvarsall'] = msvc9compiler.query_vcvarsall
        msvc9compiler.query_vcvarsall = msvc9_query_vcvarsall
    except Exception:
        pass

    try:
        # Patch distutils._msvccompiler._get_vc_env
        unpatched['msvc14_get_vc_env'] = msvc14compiler._get_vc_env
        msvc14compiler._get_vc_env = msvc14_get_vc_env
    except Exception:
        pass


def msvc9_find_vcvarsall(version):
    """
    Patched "distutils.msvc9compiler.find_vcvarsall" to use the standalone
    compiler build for Python (VCForPython). Fall back to original behavior
    when the standalone compiler is not available.

    Redirect the path of "vcvarsall.bat".

    Known supported compilers
    -------------------------
    Microsoft Visual C++ 9.0:
        Microsoft Visual C++ Compiler for Python 2.7 (x86, amd64)

    Parameters
    ----------
    version: float
        Required Microsoft Visual C++ version.

    Return
    ------
    vcvarsall.bat path: str
    """
    Reg = msvc9compiler.Reg
    VC_BASE = r'Software\%sMicrosoft\DevDiv\VCForPython\%0.1f'
    key = VC_BASE % ('', version)
    try:
        # Per-user installs register the compiler path here
        productdir = Reg.get_value(key, "installdir")
    except KeyError:
        try:
            # All-user installs on a 64-bit system register here
            key = VC_BASE % ('Wow6432Node\\', version)
            productdir = Reg.get_value(key, "installdir")
        except KeyError:
            productdir = None

    if productdir:
        vcvarsall = os.path.os.path.join(productdir, "vcvarsall.bat")
        if os.path.isfile(vcvarsall):
            return vcvarsall

    return unpatched['msvc9_find_vcvarsall'](version)


def msvc9_query_vcvarsall(ver, arch='x86', *args, **kwargs):
    """
    Patched "distutils.msvc9compiler.query_vcvarsall" for support standalones
    compilers.

    Set environment without use of "vcvarsall.bat".

    Known supported compilers
    -------------------------
    Microsoft Visual C++ 9.0:
        Microsoft Visual C++ Compiler for Python 2.7 (x86, amd64);
        Microsoft Windows SDK 7.0 (x86, x64, ia64);
        Microsoft Windows SDK 6.1 (x86, x64, ia64)

    Microsoft Visual C++ 10.0:
        Microsoft Windows SDK 7.1 (x86, x64, ia64)

    Parameters
    ----------
    ver: float
        Required Microsoft Visual C++ version.
    arch: str
        Target architecture.

    Return
    ------
    environment: dict
    """
    # Try to get environement from vcvarsall.bat (Classical way)
    try:
        return unpatched['msvc9_query_vcvarsall'](ver, arch, *args, **kwargs)
    except distutils.errors.DistutilsPlatformError:
        # Pass error if Vcvarsall.bat is missing
        pass
    except ValueError:
        # Pass error if environment not set after executing vcvarsall.bat
        pass

    # If error, try to set environment directly
    try:
        return EnvironmentInfo(arch, ver).return_env()
    except distutils.errors.DistutilsPlatformError as exc:
        _augment_exception(exc, ver, arch)
        raise


def msvc14_get_vc_env(plat_spec):
    """
    Patched "distutils._msvccompiler._get_vc_env" for support standalones
    compilers.

    Set environment without use of "vcvarsall.bat".

    Known supported compilers
    -------------------------
    Microsoft Visual C++ 14.0:
        Microsoft Visual C++ Build Tools 2015 (x86, x64, arm)

    Parameters
    ----------
    plat_spec: str
        Target architecture.

    Return
    ------
    environment: dict
    """
    # Try to get environment from vcvarsall.bat (Classical way)
    try:
        return unpatched['msvc14_get_vc_env'](plat_spec)
    except distutils.errors.DistutilsPlatformError:
        # Pass error Vcvarsall.bat is missing
        pass

    # If error, try to set environment directly
    try:
        return EnvironmentInfo(plat_spec, vc_min_ver=14.0).return_env()
    except distutils.errors.DistutilsPlatformError as exc:
        _augment_exception(exc, 14.0)
        raise


def _augment_exception(exc, version, arch=''):
    """
    Add details to the exception message to help guide the user
    as to what action will resolve it.
    """
    # Error if MSVC++ directory not found or environment not set
    message = exc.args[0]

    if "vcvarsall" in message.lower() or "visual c" in message.lower():
        # Special error message if MSVC++ not installed
        tmpl = 'Microsoft Visual C++ {version:0.1f} is required.'
        message = tmpl.format(**locals())
        msdownload = 'www.microsoft.com/download/details.aspx?id=%d'
        if version == 9.0:
            if arch.lower().find('ia64') > -1:
                # For VC++ 9.0, if IA64 support is needed, redirect user
                # to Windows SDK 7.0
                message += ' Get it with "Microsoft Windows SDK 7.0": '
                message += msdownload % 3138
            else:
                # For VC++ 9.0 redirect user to Vc++ for Python 2.7 :
                # This redirection link is maintained by Microsoft.
                # Contact vspython@microsoft.com if it needs updating.
                message += ' Get it from http://aka.ms/vcpython27'
        elif version == 10.0:
            # For VC++ 10.0 Redirect user to Windows SDK 7.1
            message += ' Get it with "Microsoft Windows SDK 7.1": '
            message += msdownload % 8279
        elif version >= 14.0:
            # For VC++ 14.0 Redirect user to Visual C++ Build Tools
            message += (' Get it with "Microsoft Visual C++ Build Tools": '
                r'http://landinghub.visualstudio.com/visual-cpp-build-tools')

    exc.args = (message, )


class PlatformInfo:
    """
    Current and Target Architectures informations.

    Parameters
    ----------
    arch: str
        Target architecture.
    """
    current_cpu = safe_env.get('processor_architecture', '').lower()

    def __init__(self, arch):
        self.arch = arch.lower().replace('x64', 'amd64')

    @property
    def target_cpu(self):
        return self.arch[self.arch.find('_') + 1:]

    def target_is_x86(self):
        return self.target_cpu == 'x86'

    def current_is_x86(self):
        return self.current_cpu == 'x86'

    def current_dir(self, hidex86=False, x64=False):
        """
        Current platform specific subfolder.

        Parameters
        ----------
        hidex86: bool
            return '' and not '\x86' if architecture is x86.
        x64: bool
            return '\x64' and not '\amd64' if architecture is amd64.

        Return
        ------
        subfolder: str
            '\target', or '' (see hidex86 parameter)
        """
        return (
            '' if (self.current_cpu == 'x86' and hidex86) else
            r'\x64' if (self.current_cpu == 'amd64' and x64) else
            r'\%s' % self.current_cpu
        )

    def target_dir(self, hidex86=False, x64=False):
        """
        Target platform specific subfolder.

        Parameters
        ----------
        hidex86: bool
            return '' and not '\x86' if architecture is x86.
        x64: bool
            return '\x64' and not '\amd64' if architecture is amd64.

        Return
        ------
        subfolder: str
            '\current', or '' (see hidex86 parameter)
        """
        return (
            '' if (self.target_cpu == 'x86' and hidex86) else
            r'\x64' if (self.target_cpu == 'amd64' and x64) else
            r'\%s' % self.target_cpu
        )

    def cross_dir(self, forcex86=False):
        """
        Cross platform specific subfolder.

        Parameters
        ----------
        forcex86: bool
            Use 'x86' as current architecture even if current acritecture is
            not x86.

        Return
        ------
        subfolder: str
            '' if target architecture is current architecture,
            '\current_target' if not.
        """
        current = 'x86' if forcex86 else self.current_cpu
        return (
            '' if self.target_cpu == current else
            self.target_dir().replace('\\', '\\%s_' % current)
        )


class RegistryInfo:
    """
    Microsoft Visual Studio related registry informations.

    Parameters
    ----------
    platform_info: PlatformInfo
        "PlatformInfo" instance.
    """
    HKEYS = (winreg.HKEY_USERS,
             winreg.HKEY_CURRENT_USER,
             winreg.HKEY_LOCAL_MACHINE,
             winreg.HKEY_CLASSES_ROOT)

    def __init__(self, platform_info):
        self.pi = platform_info

    @property
    def microsoft(self):
        """
        Microsoft software registry key.
        """
        return os.path.join(
            'Software',
            '' if self.pi.current_is_x86() else 'Wow6432Node',
            'Microsoft',
        )

    @property
    def visualstudio(self):
        """
        Microsoft Visual Studio root registry key.
        """
        return os.path.join(self.microsoft, 'VisualStudio')

    @property
    def sxs(self):
        """
        Microsoft Visual Studio SxS registry key.
        """
        return os.path.join(self.visualstudio, 'SxS')

    @property
    def vc(self):
        """
        Microsoft Visual C++ VC7 registry key.
        """
        return os.path.join(self.sxs, 'VC7')

    @property
    def vs(self):
        """
        Microsoft Visual Studio VS7 registry key.
        """
        return os.path.join(self.sxs, 'VS7')

    @property
    def vc_for_python(self):
        """
        Microsoft Visual C++ for Python registry key.
        """
        path = r'DevDiv\VCForPython'
        return os.path.join(self.microsoft, path)

    @property
    def microsoft_sdk(self):
        """
        Microsoft SDK registry key.
        """
        return os.path.join(self.microsoft, 'Microsoft SDKs')

    @property
    def windows_sdk(self):
        """
        Microsoft Windows/Platform SDK registry key.
        """
        return os.path.join(self.microsoft_sdk, 'Windows')

    @property
    def netfx_sdk(self):
        """
        Microsoft .NET Framework SDK registry key.
        """
        return os.path.join(self.microsoft_sdk, 'NETFXSDK')

    @property
    def windows_kits_roots(self):
        """
        Microsoft Windows Kits Roots registry key.
        """
        return os.path.join(self.microsoft, r'Windows Kits\Installed Roots')

    def lookup(self, key, name):
        """
        Look for values in registry.

        Parameters
        ----------
        key: str
            Registry key path where look.
        name: str
            Value name to find.

        Return
        ------
        str: value
        """
        for hkey in self.HKEYS:
            try:
                bkey = winreg.OpenKey(hkey, key, 0, winreg.KEY_READ)
            except IOError:
                continue
            try:
                return winreg.QueryValueEx(bkey, name)[0]
            except IOError:
                pass


class SystemInfo:
    """
    Microsoft Windows and Visual Studio related system inormations.

    Parameters
    ----------
    registry_info: RegistryInfo
        "RegistryInfo" instance.
    vc_ver: float
        Required Microsoft Visual C++ version.
    """
    # Variables and properties in this class use originals CamelCase variables
    # names from Microsoft source files for more easy comparaison.
    WinDir = safe_env.get('WinDir', '')
    ProgramFiles = safe_env.get('ProgramFiles', '')
    ProgramFilesx86 = safe_env.get('ProgramFiles(x86)', ProgramFiles)

    def __init__(self, registry_info, vc_ver=None):
        self.ri = registry_info
        self.pi = self.ri.pi
        if vc_ver:
            self.vc_ver = vc_ver
        else:
            try:
                self.vc_ver = self.find_available_vc_vers()[-1]
            except IndexError:
                err = 'No Microsoft Visual C++ version found'
                raise distutils.errors.DistutilsPlatformError(err)

    def find_available_vc_vers(self):
        """
        Find all available Microsoft Visual C++ versions.
        """
        vckeys = (self.ri.vc, self.ri.vc_for_python)
        vc_vers = []
        for hkey in self.ri.HKEYS:
            for key in vckeys:
                try:
                    bkey = winreg.OpenKey(hkey, key, 0, winreg.KEY_READ)
                except IOError:
                    continue
                subkeys, values, _ = winreg.QueryInfoKey(bkey)
                for i in range(values):
                    try:
                        ver = float(winreg.EnumValue(bkey, i)[0])
                        if ver not in vc_vers:
                            vc_vers.append(ver)
                    except ValueError:
                        pass
                for i in range(subkeys):
                    try:
                        ver = float(winreg.EnumKey(bkey, i))
                        if ver not in vc_vers:
                            vc_vers.append(ver)
                    except ValueError:
                        pass
        return sorted(vc_vers)

    @property
    def VSInstallDir(self):
        """
        Microsoft Visual Studio directory.
        """
        # Default path
        name = 'Microsoft Visual Studio %0.1f' % self.vc_ver
        default = os.path.join(self.ProgramFilesx86, name)

        # Try to get path from registry, if fail use default path
        return self.ri.lookup(self.ri.vs, '%0.1f' % self.vc_ver) or default

    @property
    def VCInstallDir(self):
        """
        Microsoft Visual C++ directory.
        """
        # Default path
        default = r'Microsoft Visual Studio %0.1f\VC' % self.vc_ver
        guess_vc = os.path.join(self.ProgramFilesx86, default)

        # Try to get "VC++ for Python" path from registry as default path
        reg_path = os.path.join(self.ri.vc_for_python, '%0.1f' % self.vc_ver)
        python_vc = self.ri.lookup(reg_path, 'installdir')
        default_vc = os.path.join(python_vc, 'VC') if python_vc else guess_vc

        # Try to get path from registry, if fail use default path
        path = self.ri.lookup(self.ri.vc, '%0.1f' % self.vc_ver) or default_vc

        if not os.path.isdir(path):
            msg = 'Microsoft Visual C++ directory not found'
            raise distutils.errors.DistutilsPlatformError(msg)

        return path

    @property
    def WindowsSdkVersion(self):
        """
        Microsoft Windows SDK versions.
        """
        # Set Windows SDK versions for specified MSVC++ version
        if self.vc_ver <= 9.0:
            return ('7.0', '6.1', '6.0a')
        elif self.vc_ver == 10.0:
            return ('7.1', '7.0a')
        elif self.vc_ver == 11.0:
            return ('8.0', '8.0a')
        elif self.vc_ver == 12.0:
            return ('8.1', '8.1a')
        elif self.vc_ver >= 14.0:
            return ('10.0', '8.1')

    @property
    def WindowsSdkDir(self):
        """
        Microsoft Windows SDK directory.
        """
        sdkdir = ''
        for ver in self.WindowsSdkVersion:
            # Try to get it from registry
            loc = os.path.join(self.ri.windows_sdk, 'v%s' % ver)
            sdkdir = self.ri.lookup(loc, 'installationfolder')
            if sdkdir:
                break
        if not sdkdir or not os.path.isdir(sdkdir):
            # Try to get "VC++ for Python" version from registry
            path = os.path.join(self.ri.vc_for_python, '%0.1f' % self.vc_ver)
            install_base = self.ri.lookup(path, 'installdir')
            if install_base:
                sdkdir = os.path.join(install_base, 'WinSDK')
        if not sdkdir or not os.path.isdir(sdkdir):
            # If fail, use default new path
            for ver in self.WindowsSdkVersion:
                intver = ver[:ver.rfind('.')]
                path = r'Microsoft SDKs\Windows Kits\%s' % (intver)
                d = os.path.join(self.ProgramFiles, path)
                if os.path.isdir(d):
                    sdkdir = d
        if not sdkdir or not os.path.isdir(sdkdir):
            # If fail, use default old path
            for ver in self.WindowsSdkVersion:
                path = r'Microsoft SDKs\Windows\v%s' % ver
                d = os.path.join(self.ProgramFiles, path)
                if os.path.isdir(d):
                    sdkdir = d
        if not sdkdir:
            # If fail, use Platform SDK
            sdkdir = os.path.join(self.VCInstallDir, 'PlatformSDK')
        return sdkdir

    @property
    def WindowsSDKExecutablePath(self):
        """
        Microsoft Windows SDK executable directory.
        """
        # Find WinSDK NetFx Tools registry dir name
        if self.vc_ver <= 11.0:
            netfxver = 35
            arch = ''
        else:
            netfxver = 40
            hidex86 = True if self.vc_ver <= 12.0 else False
            arch = self.pi.current_dir(x64=True, hidex86=hidex86)
        fx = 'WinSDK-NetFx%dTools%s' % (netfxver, arch.replace('\\', '-'))

        # liste all possibles registry paths
        regpaths = []
        if self.vc_ver >= 14.0:
            for ver in self.NetFxSdkVersion:
                regpaths += [os.path.join(self.ri.netfx_sdk, ver, fx)]

        for ver in self.WindowsSdkVersion:
            regpaths += [os.path.join(self.ri.windows_sdk, 'v%sA' % ver, fx)]

        # Return installation folder from the more recent path
        for path in regpaths:
            execpath = self.ri.lookup(path, 'installationfolder')
            if execpath:
                break
        return execpath

    @property
    def FSharpInstallDir(self):
        """
        Microsoft Visual F# directory.
        """
        path = r'%0.1f\Setup\F#' % self.vc_ver
        path = os.path.join(self.ri.visualstudio, path)
        return self.ri.lookup(path, 'productdir') or ''

    @property
    def UniversalCRTSdkDir(self):
        """
        Microsoft Universal CRT SDK directory.
        """
        # Set Kit Roots versions for specified MSVC++ version
        if self.vc_ver >= 14.0:
            vers = ('10', '81')
        else:
            vers = ()

        # Find path of the more recent Kit
        for ver in vers:
            sdkdir = self.ri.lookup(self.ri.windows_kits_roots,
                                    'kitsroot%s' % ver)
            if sdkdir:
                break
        return sdkdir or ''

    @property
    def NetFxSdkVersion(self):
        """
        Microsoft .NET Framework SDK versions.
        """
        # Set FxSdk versions for specified MSVC++ version
        if self.vc_ver >= 14.0:
            return ('4.6.1', '4.6')
        else:
            return ()

    @property
    def NetFxSdkDir(self):
        """
        Microsoft .NET Framework SDK directory.
        """
        for ver in self.NetFxSdkVersion:
            loc = os.path.join(self.ri.netfx_sdk, ver)
            sdkdir = self.ri.lookup(loc, 'kitsinstallationfolder')
            if sdkdir:
                break
        return sdkdir or ''

    @property
    def FrameworkDir32(self):
        """
        Microsoft .NET Framework 32bit directory.
        """
        # Default path
        guess_fw = os.path.join(self.WinDir, r'Microsoft.NET\Framework')

        # Try to get path from registry, if fail use default path
        return self.ri.lookup(self.ri.vc, 'frameworkdir32') or guess_fw

    @property
    def FrameworkDir64(self):
        """
        Microsoft .NET Framework 64bit directory.
        """
        # Default path
        guess_fw = os.path.join(self.WinDir, r'Microsoft.NET\Framework64')

        # Try to get path from registry, if fail use default path
        return self.ri.lookup(self.ri.vc, 'frameworkdir64') or guess_fw

    @property
    def FrameworkVersion32(self):
        """
        Microsoft .NET Framework 32bit versions.
        """
        return self._find_dot_net_versions(32)

    @property
    def FrameworkVersion64(self):
        """
        Microsoft .NET Framework 64bit versions.
        """
        return self._find_dot_net_versions(64)

    def _find_dot_net_versions(self, bits=32):
        """
        Find Microsoft .NET Framework versions.

        Parameters
        ----------
        bits: int
            Platform number of bits: 32 or 64.
        """
        # Find actual .NET version
        ver = self.ri.lookup(self.ri.vc, 'frameworkver%d' % bits) or ''

        # Set .NET versions for specified MSVC++ version
        if self.vc_ver >= 12.0:
            frameworkver = (ver, 'v4.0')
        elif self.vc_ver >= 10.0:
            frameworkver = ('v4.0.30319' if ver.lower()[:2] != 'v4' else ver,
                            'v3.5')
        elif self.vc_ver == 9.0:
            frameworkver = ('v3.5', 'v2.0.50727')
        if self.vc_ver == 8.0:
            frameworkver = ('v3.0', 'v2.0.50727')
        return frameworkver


class EnvironmentInfo:
    """
    Return environment variables for specified Microsoft Visual C++ version
    and platform : Lib, Include, Path and libpath.

    This function is compatible with Microsoft Visual C++ 9.0 to 14.0.

    Script created by analysing Microsoft environment configuration files like
    "vcvars[...].bat", "SetEnv.Cmd", "vcbuildtools.bat", ...

    Parameters
    ----------
    arch: str
        Target architecture.
    vc_ver: float
        Required Microsoft Visual C++ version. If not set, autodetect the last
        version.
    vc_min_ver: float
        Minimum Microsoft Visual C++ version.
    """
    # Variables and properties in this class use originals CamelCase variables
    # names from Microsoft source files for more easy comparaison.

    def __init__(self, arch, vc_ver=None, vc_min_ver=None):
        self.pi = PlatformInfo(arch)
        self.ri = RegistryInfo(self.pi)
        self.si = SystemInfo(self.ri, vc_ver)

        if vc_min_ver:
            if self.vc_ver < vc_min_ver:
                err = 'No suitable Microsoft Visual C++ version found'
                raise distutils.errors.DistutilsPlatformError(err)

    @property
    def vc_ver(self):
        """
        Microsoft Visual C++ version.
        """
        return self.si.vc_ver

    @property
    def VSTools(self):
        """
        Microsoft Visual Studio Tools
        """
        paths = [r'Common7\IDE', r'Common7\Tools']

        if self.vc_ver >= 14.0:
            arch_subdir = self.pi.current_dir(hidex86=True, x64=True)
            paths += [r'Common7\IDE\CommonExtensions\Microsoft\TestWindow']
            paths += [r'Team Tools\Performance Tools']
            paths += [r'Team Tools\Performance Tools%s' % arch_subdir]

        return [os.path.join(self.si.VSInstallDir, path) for path in paths]

    @property
    def VCIncludes(self):
        """
        Microsoft Visual C++ & Microsoft Foundation Class Includes
        """
        return [os.path.join(self.si.VCInstallDir, 'Include'),
                os.path.join(self.si.VCInstallDir, 'ATLMFC\Include')]

    @property
    def VCLibraries(self):
        """
        Microsoft Visual C++ & Microsoft Foundation Class Libraries
        """
        arch_subdir = self.pi.target_dir(hidex86=True)
        paths = ['Lib%s' % arch_subdir, r'ATLMFC\Lib%s' % arch_subdir]

        if self.vc_ver >= 14.0:
            paths += [r'Lib\store%s' % arch_subdir]

        return [os.path.join(self.si.VCInstallDir, path) for path in paths]

    @property
    def VCStoreRefs(self):
        """
        Microsoft Visual C++ store references Libraries
        """
        if self.vc_ver < 14.0:
            return []
        return [os.path.join(self.si.VCInstallDir, r'Lib\store\references')]

    @property
    def VCTools(self):
        """
        Microsoft Visual C++ Tools
        """
        si = self.si
        tools = [os.path.join(si.VCInstallDir, 'VCPackages')]

        forcex86 = True if self.vc_ver <= 10.0 else False
        arch_subdir = self.pi.cross_dir(forcex86)
        if arch_subdir:
            tools += [os.path.join(si.VCInstallDir, 'Bin%s' % arch_subdir)]

        if self.vc_ver >= 14.0:
            path = 'Bin%s' % self.pi.current_dir(hidex86=True)
            tools += [os.path.join(si.VCInstallDir, path)]

        else:
            tools += [os.path.join(si.VCInstallDir, 'Bin')]

        return tools

    @property
    def OSLibraries(self):
        """
        Microsoft Windows SDK Libraries
        """
        if self.vc_ver <= 10.0:
            arch_subdir = self.pi.target_dir(hidex86=True, x64=True)
            return [os.path.join(self.si.WindowsSdkDir, 'Lib%s' % arch_subdir)]

        else:
            arch_subdir = self.pi.target_dir(x64=True)
            lib = os.path.join(self.si.WindowsSdkDir, 'lib')
            libver = self._get_content_dirname(lib)
            return [os.path.join(lib, '%sum%s' % (libver, arch_subdir))]

    @property
    def OSIncludes(self):
        """
        Microsoft Windows SDK Include
        """
        include = os.path.join(self.si.WindowsSdkDir, 'include')

        if self.vc_ver <= 10.0:
            return [include, os.path.join(include, 'gl')]

        else:
            if self.vc_ver >= 14.0:
                sdkver = self._get_content_dirname(include)
            else:
                sdkver = ''
            return [os.path.join(include, '%sshared' % sdkver),
                    os.path.join(include, '%sum' % sdkver),
                    os.path.join(include, '%swinrt' % sdkver)]

    @property
    def OSLibpath(self):
        """
        Microsoft Windows SDK Libraries Paths
        """
        ref = os.path.join(self.si.WindowsSdkDir, 'References')
        libpath = []

        if self.vc_ver <= 9.0:
            libpath += self.OSLibraries

        if self.vc_ver >= 11.0:
            libpath += [os.path.join(ref, r'CommonConfiguration\Neutral')]

        if self.vc_ver >= 14.0:
            libpath += [
                ref,
                os.path.join(self.si.WindowsSdkDir, 'UnionMetadata'),
                os.path.join(
                    ref,
                    'Windows.Foundation.UniversalApiContract'
                    '1.0.0.0',
                ),
                os.path.join(
                    ref,
                    'Windows.Foundation.FoundationContract',
                    '1.0.0.0',
                ),
                os.path.join(
                    ref,
                    'Windows.Networking.Connectivity.WwanContract'
                    '1.0.0.0',
                ),
                os.path.join(
                    self.si.WindowsSdkDir,
                    'ExtensionSDKs',
                    'Microsoft.VCLibs',
                    '%0.1f' % self.vc_ver,
                    'References',
                    'CommonConfiguration',
                    'neutral',
                ),
            ]
        return libpath

    @property
    def SdkTools(self):
        """
        Microsoft Windows SDK Tools
        """
        bin_dir = 'Bin' if self.vc_ver <= 11.0 else r'Bin\x86'
        tools = [os.path.join(self.si.WindowsSdkDir, bin_dir)]

        if not self.pi.current_is_x86():
            arch_subdir = self.pi.current_dir(x64=True)
            path = 'Bin%s' % arch_subdir
            tools += [os.path.join(self.si.WindowsSdkDir, path)]

        if self.vc_ver == 10.0 or self.vc_ver == 11.0:
            if self.pi.target_is_x86():
                arch_subdir = ''
            else:
                arch_subdir = self.pi.current_dir(hidex86=True, x64=True)
            path = r'Bin\NETFX 4.0 Tools%s' % arch_subdir
            tools += [os.path.join(self.si.WindowsSdkDir, path)]

        if self.si.WindowsSDKExecutablePath:
            tools += [self.si.WindowsSDKExecutablePath]

        return tools

    @property
    def SdkSetup(self):
        """
        Microsoft Windows SDK Setup
        """
        if self.vc_ver > 9.0:
            return []

        return [os.path.join(self.si.WindowsSdkDir, 'Setup')]

    @property
    def FxTools(self):
        """
        Microsoft .NET Framework Tools
        """
        pi = self.pi
        si = self.si

        if self.vc_ver <= 10.0:
            include32 = True
            include64 = not pi.target_is_x86() and not pi.current_is_x86()
        else:
            include32 = pi.target_is_x86() or pi.current_is_x86()
            include64 = pi.current_cpu == 'amd64' or pi.target_cpu == 'amd64'

        tools = []
        if include32:
            tools += [os.path.join(si.FrameworkDir32, ver)
                      for ver in si.FrameworkVersion32]
        if include64:
            tools += [os.path.join(si.FrameworkDir64, ver)
                      for ver in si.FrameworkVersion64]
        return tools

    @property
    def NetFxSDKLibraries(self):
        """
        Microsoft .Net Framework SDK Libraries
        """
        if self.vc_ver < 14.0 or not self.si.NetFxSdkDir:
            return []

        arch_subdir = self.pi.target_dir(x64=True)
        return [os.path.join(self.si.NetFxSdkDir, r'lib\um%s' % arch_subdir)]

    @property
    def NetFxSDKIncludes(self):
        """
        Microsoft .Net Framework SDK Includes
        """
        if self.vc_ver < 14.0 or not self.si.NetFxSdkDir:
            return []

        return [os.path.join(self.si.NetFxSdkDir, r'include\um')]

    @property
    def VsTDb(self):
        """
        Microsoft Visual Studio Team System Database
        """
        return [os.path.join(self.si.VSInstallDir, r'VSTSDB\Deploy')]

    @property
    def MSBuild(self):
        """
        Microsoft Build Engine
        """
        if self.vc_ver < 12.0:
            return []

        arch_subdir = self.pi.current_dir(hidex86=True)
        path = r'MSBuild\%0.1f\bin%s' % (self.vc_ver, arch_subdir)
        return [os.path.join(self.si.ProgramFilesx86, path)]

    @property
    def HTMLHelpWorkshop(self):
        """
        Microsoft HTML Help Workshop
        """
        if self.vc_ver < 11.0:
            return []

        return [os.path.join(self.si.ProgramFilesx86, 'HTML Help Workshop')]

    @property
    def UCRTLibraries(self):
        """
        Microsoft Universal CRT Libraries
        """
        if self.vc_ver < 14.0:
            return []

        arch_subdir = self.pi.target_dir(x64=True)
        lib = os.path.join(self.si.UniversalCRTSdkDir, 'lib')
        ucrtver = self._get_content_dirname(lib)
        return [os.path.join(lib, '%sucrt%s' % (ucrtver, arch_subdir))]

    @property
    def UCRTIncludes(self):
        """
        Microsoft Universal CRT Include
        """
        if self.vc_ver < 14.0:
            return []

        include = os.path.join(self.si.UniversalCRTSdkDir, 'include')
        ucrtver = self._get_content_dirname(include)
        return [os.path.join(include, '%sucrt' % ucrtver)]

    @property
    def FSharp(self):
        """
        Microsoft Visual F#
        """
        if self.vc_ver < 11.0 and self.vc_ver > 12.0:
            return []

        return self.si.FSharpInstallDir

    @property
    def VCRuntimeRedist(self):
        """
        Microsoft Visual C++ runtime redistribuable dll
        """
        arch_subdir = self.pi.target_dir(x64=True)
        vcruntime = 'redist%s\\Microsoft.VC%d0.CRT\\vcruntime%d0.dll'
        vcruntime = vcruntime % (arch_subdir, self.vc_ver, self.vc_ver)
        return os.path.join(self.si.VCInstallDir, vcruntime)

    def return_env(self, exists=True):
        """
        Return environment dict.

        Parameters
        ----------
        exists: bool
            It True, only return existing paths.
        """
        env = dict(
            include=self._build_paths('include',
                                      [self.VCIncludes,
                                       self.OSIncludes,
                                       self.UCRTIncludes,
                                       self.NetFxSDKIncludes],
                                      exists),
            lib=self._build_paths('lib',
                                  [self.VCLibraries,
                                   self.OSLibraries,
                                   self.FxTools,
                                   self.UCRTLibraries,
                                   self.NetFxSDKLibraries],
                                  exists),
            libpath=self._build_paths('libpath',
                                      [self.VCLibraries,
                                       self.FxTools,
                                       self.VCStoreRefs,
                                       self.OSLibpath],
                                      exists),
            path=self._build_paths('path',
                                   [self.VCTools,
                                    self.VSTools,
                                    self.VsTDb,
                                    self.SdkTools,
                                    self.SdkSetup,
                                    self.FxTools,
                                    self.MSBuild,
                                    self.HTMLHelpWorkshop,
                                    self.FSharp],
                                   exists),
        )
        if self.vc_ver >= 14 and os.path.isfile(self.VCRuntimeRedist):
            env['py_vcruntime_redist'] = self.VCRuntimeRedist
        return env

    def _build_paths(self, name, spec_path_lists, exists):
        """
        Given an environment variable name and specified paths,
        return a pathsep-separated string of paths containing
        unique, extant, directories from those paths and from
        the environment variable. Raise an error if no paths
        are resolved.
        """
        # flatten spec_path_lists
        spec_paths = itertools.chain.from_iterable(spec_path_lists)
        env_paths = safe_env.get(name, '').split(os.pathsep)
        paths = itertools.chain(spec_paths, env_paths)
        extant_paths = list(filter(os.path.isdir, paths)) if exists else paths
        if not extant_paths:
            msg = "%s environment variable is empty" % name.upper()
            raise distutils.errors.DistutilsPlatformError(msg)
        unique_paths = self._unique_everseen(extant_paths)
        return os.pathsep.join(unique_paths)

    # from Python docs
    def _unique_everseen(self, iterable, key=None):
        """
        List unique elements, preserving order.
        Remember all elements ever seen.

        _unique_everseen('AAAABBBCCDAABBB') --> A B C D

        _unique_everseen('ABBCcAD', str.lower) --> A B C D
        """
        seen = set()
        seen_add = seen.add
        if key is None:
            for element in filterfalse(seen.__contains__, iterable):
                seen_add(element)
                yield element
        else:
            for element in iterable:
                k = key(element)
                if k not in seen:
                    seen_add(k)
                    yield element

    def _get_content_dirname(self, path):
        """
        Return name of the first dir in path or '' if no dir found.

        Parameters
        ----------
        path: str
            Path where search dir.

        Return
        ------
        foldername: str
            "name\" or ""
        """
        try:
            name = os.listdir(path)
            if name:
                return '%s\\' % name[0]
            return ''
        except IOError:
            return ''



"""
Learns trips
"""
import numpy as np
from .similarity import segment_similarity

def complete_trip(canonical_trips, from_point, to_point):
    """ Completes a trip based on set of canonical trips

    Args:
        canonical_trips (:obj:`list` of :obj:`Segment`)
        from_point (:obj:`Point`)
        to_point (:obj:`Point`)
    """
    result = []
    weights = []
    total_weights = 0.0
    # match points in lines
    for (_, trip, count) in canonical_trips:
        from_index = trip.closest_point_to(from_point)
        to_index = trip.closest_point_to(to_point)
        if from_index != to_index and from_index != -1 and to_index != -1:
            trip_slice = trip.slice(from_index, to_index)
            result.append([[p.lat, p.lon] for p in trip_slice.points])
            weights.append(count)
            total_weights = total_weights + count

    return {
        'possibilities': result,
        'weights': list(np.array(weights) / total_weights)
    }


def learn_trip(current, current_id, canonical_trips, insert_canonical, update_canonical, eps):
    """Learns a trip against of other canonical trips

    Args:
        current (:obj:`Track`): current trip
        current_id (int): current trip (db) id
        canonical_trips (:obj:`list` of :obj:`Track`): list of canonical trips
        insert_canonical: Function to insert a new canonical trip, with the
            signature: (:obj:`Track`, int) -> void
        update_canonical: Function to update an existing canonical trip, with
            the signature: (int, :obj:`Track`, int) -> void
    """

    if len(canonical_trips) == 0:
        current.simplify(eps, 0, 0, topology_only=True)
        insert_canonical(current, current_id)
    else:
        canonical_trips = [
            (trip_id, trip, segment_similarity(trip, current))
            for trip_id, trip in canonical_trips
            ]
        canonical_trips = sorted(canonical_trips, key=lambda t: t[2][0])

        trip_id, trip, (similarity, _) = canonical_trips[0]

        if similarity >= 0.8:
            # Same trip, fit all segments
            trip.merge_and_fit(current)#, diffs)
            trip.simplify(eps, 0, topology_only=True)
            update_canonical(trip_id, trip, current_id)

        elif similarity >= 0.3:
            # Fit similar segments
            orig = trip.copy()
            trip.merge_and_fit(current)#, diffs, threshold=0.2)
            trip.simplify(eps, 0, topology_only=True)

            current.merge_and_fit(orig)#, diffs, threshold=0.8)
            current.simplify(eps, 0, topology_only=True)

            # updateCanonicalTrip(tripId, trip, currentTripId)
            # updateCanonicalTrip(tripId, trip, currentTripId)

        else:
            # Insert new canonical representation
            current.simplify(eps, 0, topology_only=True)
            insert_canonical(current, current_id)



"""
Segment preprocessing
"""
import datetime
from .point import Point

MIN_TIME = datetime.datetime(1999, 1, 1)

def preprocess_segment(points, max_acc, destructive=True):
    """ Calculates the metrics of points and unrealistic ones

    Uses their ordering to calculate velocities and spatio-temporal distances
    If destructive it removes points with accelarations higher than max_acc
        and that occured before MIN_TIME

    Args:
        points (:obj:`list` of :obj:`Point`)
        max_acc (float): Maximum acceleration threshold
        destructive (bool, optional): True to allow the removal of elements
            from the initial array. Default is True
    Returns:
        A tuple with the resulting array of points and the
        the points dropped
    """
    result = []
    last = None
    for point in points:
        if not (destructive and point.time < MIN_TIME):
            current = Point(point.lat, point.lon, point.time)

            if last is not None:
                current.compute_metrics(last)

            if destructive and abs(current.acc) > max_acc:
                current = Point(last.lat, last.lon, point.time)
                current.compute_metrics(last)

            result.append(current)
            last = current

    return result



"""
Learns trips
"""
from .similarity import segment_similarity

def learn_trip(current, current_id, canonical_trips, insert_canonical, update_canonical):
    """Learns a trip against of other canonical trips

    Args:
        current (:obj:`Track`): current trip
        current_id (int): current trip (db) id
        canonical_trips (:obj:`list` of :obj:`Track`): list of canonical trips
        insert_canonical: Function to insert a new canonical trip, with the
            signature: (:obj:`Track`, int) -> void
        update_canonical: Function to update an existing canonical trip, with
            the signature: (int, :obj:`Track`, int) -> void
    """

    if len(canonical_trips) == 0:
        # print("Insert new canonical trip, unmatched bounding box")

        # TODO simplify
        # currentTrip.simplify(topology_only=True)
        # print(currentTrip, currentTripId)
        insert_canonical(current, current_id)
    else:
        # print(canonicalTrips)
        canonical_trips = [
            (trip_id, trip, segment_similarity(trip, current))
            for trip_id, trip in canonical_trips
            ]
        canonical_trips = sorted(canonical_trips, key=lambda t: t[2][0])

        trip_id, trip, (similarity, _) = canonical_trips[0]
        # print(similarity, diffs)

        if similarity >= 0.8:
            # print("Same as canonical trip, fitting trip")
            # Same trip, fit all segments
            trip.merge_and_fit(current)#, diffs)
            trip.simplify(topology_only=True)
            update_canonical(trip_id, trip, current_id)

        elif similarity >= 0.3:
            # print("Similar as canonical trip, fitting similar segments. Adding new trip")
            # Fit similar segments
            orig = trip.copy()
            trip.merge_and_fit(current)#, diffs, threshold=0.2)
            trip.simplify(topology_only=True)

            current.merge_and_fit(orig)#, diffs, threshold=0.8)
            current.simplify(topology_only=True)

            # updateCanonicalTrip(tripId, trip, currentTripId)
            # updateCanonicalTrip(tripId, trip, currentTripId)

        else:
            # print("New canonical trip, with similar boundingbox to others")
            # Insert new canonical representation
            current.simplify(topology_only=True)
            insert_canonical(current, current_id)



"""
Smoothing methods
"""
import copy
import numpy as np
from .point import Point
from .kalman import kalman_filter

INVERSE_STRATEGY = 0
EXTRAPOLATE_STRATEGY = 1

def extrapolate_points(points, n_points):
    """ Extrapolate a number of points, based on the first ones

    Args:
        points (:obj:`list` of :obj:`Point`)
        n_points (int): number of points to extrapolate
    Returns:
        :obj:`list` of :obj:`Point`
    """
    points = points[:]
    lat = []
    lon = []
    last = None
    for point in points:
        if last is not None:
            lat.append(last.lat-point.lat)
            lon.append(last.lon-point.lon)
        last = point

    dts = np.mean([p.dt for p in points])
    lons = np.mean(lon)
    lats = np.mean(lat)

    gen_sample = []
    last = points[0]
    for _ in range(n_points):
        point = Point(last.lat+lats, last.lon+lons, last.time - dts)
        point.compute_metrics(last)
        gen_sample.append(point)
        last = point

    return gen_sample

def with_extrapolation(points, noise, n_points):
    """ Smooths a set of points, but it extrapolates some points at the beginning

    Args:
        points (:obj:`list` of :obj:`Point`)
        noise (float): Expected noise, the higher it is the more the path will
            be smoothed.
    Returns:
        :obj:`list` of :obj:`Point`
    """
    return kalman_filter(extrapolate_points(points, n_points) + points, noise)[n_points:]

def with_inverse(points, noise):
    """ Smooths a set of points

    It smooths them twice, once in given order, another one in the reverse order.
    The the first half of the results will be taken from the reverse order and
        the second half from the normal order.

    Args:
        points (:obj:`list` of :obj:`Point`)
        noise (float): Expected noise, the higher it is the more the path will
            be smoothed.
    Returns:
        :obj:`list` of :obj:`Point`
    """
    noise_sample = 20
    n_points = len(points)/2
    break_point = n_points - noise_sample

    points_part = copy.deepcopy(points[:n_points])
    points_part = list(reversed(points_part))
    part = kalman_filter(points_part, noise)
    total = kalman_filter(points, noise)

    return list(reversed(part))[:break_point] + total[break_point:]



"""
Track class
"""
from copy import deepcopy
from os.path import basename

import gpxpy
import numpy as np
from rtree import index

from .segment import Segment
from .similarity import segment_similarity

DEFAULT_FILE_NAME_FORMAT = "%Y-%m-%d"

class Track(object):
    """Collection of segments

    This is a higher level class, all methods of TrackToTrip library
    can be called over this class

    Attributes:
        name: A string indicating the name of the track
        segments: Array of TrackToTrip Segments
        preprocessed: Boolean, true if it has been preprocessed
    """

    def __init__(self, name, segments):
        """ Constructor

        When constructing a track it's not guaranteed that the segments
        have their properties computed. Call preprocess method over this
        class, or over each segment to guarantee it.

        Args:
            name (:obj:`str`)
            segments(:obj:`list` of :obj:`Segment`)
        """
        self.name = name
        self.segments = sorted(segments, key=lambda s: s.points[0].time)
        self.preprocessed = False

    # def start_time(self):
    #     lastTime = None
    #     for segment in self.segments:
    #         if lastTime is None:
    #             lastTime = segment.getStartTime()
    #         elif lastTime > segment.getStartTime():
    #             lastTime = segment.getStartTime()
    #     return lastTime

    def generate_name(self, name_format=DEFAULT_FILE_NAME_FORMAT):
        """ Generates a name for the track

        The name is generated based on the date of the first point of the
        track, or in case it doesn't exist, "EmptyTrack"

        Args:
            name_format (str, optional): Name formar to give to the track, based on
                its start time. Defaults to DEFAULT_FILE_NAME_FORMAT
        Returns:
            str
        """
        if len(self.segments) > 0:
            return self.segments[0].points[0].time.strftime(name_format) + ".gpx"
        else:
            return "EmptyTrack"
    #
    # def remove_noise(self, var=2):
    #     """In-place removal of noise points
    #
    #     Arguments:
    #         var: Number to adjust noise removal sensitivity
    #     Returns:
    #         This track
    #     """
    #     for segment in self.segments:
    #         segment.removeNoise(var)
    #     return self

    def smooth(self, strategy, noise):
        """In-place smoothing of segments

        Returns:
            This track
        """
        for segment in self.segments:
            segment.smooth(strategy, noise)
        return self

    def segment(self, eps, min_time):
        """In-place segmentation of segments

        Spatio-temporal segmentation of each segment
        The number of segments may increse after this step

        Returns:
            This track
        """
        new_segments = []
        for segment in self.segments:
            segmented = segment.segment(eps, min_time)
            for seg in segmented:
                new_segments.append(Segment(seg))
        self.segments = new_segments
        return self

    def simplify(self, eps, dist_threshold, topology_only=False):
        """In-place simplification of segments

        Args:
            topology_only: Boolean, optional. True to keep
                the topology, neglecting velocity and time
                accuracy (use common Douglas-Ramen-Peucker).
                False (default) to simplify segments keeping
                the velocity between points.
        Returns:
            This track
        """
        for segment in self.segments:
            segment.simplify(eps, dist_threshold, topology_only)
        return self

    def infer_transportation_mode(self, clf, min_time):
        """In-place transportation mode inferring of segments

        Returns:
            This track
        """
        for segment in self.segments:
            segment.infer_transportation_mode(clf, min_time)
        return self

    def copy(self):
        """Creates a deep copy of itself

        All segments and points are copied

        Returns:
            A Track object different from this instance
        """
        return deepcopy(self)

    def to_trip(
            self,
            name,
            # noise_var=2,
            smooth_strategy,
            smooth_noise,
            seg_eps,
            seg_min_time,
            simplify_dist_threshold,
            # simplify_max_time=5,
            file_format
        ):
        """In-place, transformation of a track into a trip

        A trip is a more accurate depiction of reality than a
        track.
        For a track to become a trip it need to go through the
        following steps:
            + noise removal
            + smoothing
            + spatio-temporal segmentation
            + simplification
        At the end of these steps we have a less noisy, track
        that has less points, but that holds the same information.
        It's required that each segment has their metrics calculated
        or has been preprocessed.

        Args:
            name: An optional string with the name of the trip. If
                none is given, one will be generated by generateName
        Returns:
            This Track instance
        """
        if len(name) != 0:
            name = self.name
        else:
            name = self.generate_name(file_format)

        # self.removeNoise(noise_var)

        self.smooth(smooth_strategy, smooth_noise)

        self.compute_metrics()
        self.segment(seg_eps, seg_min_time)

        self.compute_metrics()
        self.simplify(None, simplify_dist_threshold)

        self.compute_metrics()
        self.name = name

        return self

    def preprocess(self, max_acc, destructive=True):
        """In-place preprocessing of segments

        Args:
            destructive: Optional, boolean. True to allow point
                removal. More details in preprocessSegment
        Returns:
            This track
        """
        self.segments = [segment.preprocess(max_acc, destructive) for segment in self.segments]
        self.preprocessed = True
        return self

    def infer_transportation_modes(self, dt_threshold=10):
        """In-place transportation inferring of segments

        Returns:
            This track
        """
        self.segments = [
            segment.infer_transportation_mode(dt_threshold=dt_threshold)
            for segment in self.segments
            ]
        return self

    # TODO
    def infer_location(
            self,
            location_query,
            max_distance,
            google_key,
            limit
        ):
        """In-place location inferring of segments

        Returns:
            This track
        """
        self.segments = [
            segment.infer_location(location_query, max_distance, google_key, limit)
            for segment in self.segments
            ]
        return self

    def to_json(self):
        """Converts track to a JSON serializable format

        Returns:
            Map with the name, and segments of the track.
        """
        return {
            'name': self.name,
            'segments': [segment.to_json() for segment in self.segments]
            }

    # TODO
    def merge_and_fit(self, track, pairings):
        """ Merges another track with this one, ordering the points based on a
            distance heuristic

        Args:
            track (:obj:`Track`): Track to merge with
            pairings
        Returns:
            :obj:`Segment`: self
        """
        for (self_seg_index, track_seg_index, _) in pairings:
            self_s = self.segments[self_seg_index]
            track_s = track.segments[track_seg_index]

            self_s.merge_and_fit(track_s)
        return self

    def get_point_index(self, point):
        """ Gets of the closest first point

        Args:
            point (:obj:`Point`)
        Returns:
            (int, int): Segment id and point index in that segment
        """
        for i, segment in enumerate(self.segments):
            idx = segment.getPointIndex(point)
            if idx != -1:
                return i, idx
        return -1, -1

    def bounds(self):
        """ Gets the bounds of this segment

        Returns:
            (float, float, float, float): Bounds, with min latitude, min longitude,
                max latitude and max longitude
        """
        min_lat = float("inf")
        min_lon = float("inf")
        max_lat = -float("inf")
        max_lon = -float("inf")
        for segment in self.segments:
            milat, milon, malat, malon = segment.getBounds()
            min_lat = min(milat, min_lat)
            min_lon = min(milon, min_lon)
            max_lat = max(malat, max_lat)
            max_lon = max(malon, max_lon)
        return min_lat, min_lon, max_lat, max_lon

    def has_point(self, point):
        """ Checks if a point exist in any of the segments

        Args:
            points (:obj:`Point`)
        Returns:
            bool
        """
        s_ix, _ = self.get_point_index(point)
        return s_ix != -1

    def similarity(self, track):
        """ Compares two tracks based on their topology

        This method compares the given track against this
        instance. It only verifies if given track is close
        to this one, not the other way arround

        Args:
            track (:obj:`Track`)
        Returns:
            Two-tuple with global similarity between tracks
            and an array the similarity between segments
        """
        idx = index.Index()
        i = 0
        for i, segment in enumerate(self.segments):
            idx.insert(i, segment.bounds(), obj=segment)

        final_siml = []
        final_diff = []
        for i, segment in enumerate(track.segments):
            query = idx.intersection(segment.bounds(), objects=True)

            res_siml = []
            res_diff = []
            for result in query:
                siml, diff = segment_similarity(segment, result.object)
                res_siml.append(siml)
                res_diff.append((result.id, i, diff))
                # print(result.id, i, diff)

            # print("seg match", res_siml, res_diff)

            if len(res_siml) > 0:
                final_siml.append(max(res_siml))
                final_diff.append(res_diff[np.argmax(res_siml)])
            else:
                final_siml.append(0)
                final_diff.append([])

        return np.mean(final_siml), final_diff

    def compute_metrics(self):
        """ Computes metrics for every segment's point

        See Segment.compute_metrics

        Returns:
            :obj:`Track`: Self
        """
        for segment in self.segments:
            segment.compute_metrics()
        return self

    def to_gpx(self):
        """Converts track to a GPX format

        Uses GPXPY library as an intermediate format

        Returns:
            A string with the GPX/XML track
        """
        gpx = gpxpy.gpx.GPX()
        gpx_track = gpxpy.gpx.GPXTrack()
        gpx.tracks.append(gpx_track)

        for segment in self.segments:
            gpx_segment = gpxpy.gpx.GPXTrackSegment()
            gpx_track.segments.append(gpx_segment)

            for point in segment.points:
                gpx_point = gpxpy.gpx.GPXTrackPoint(point.lat, point.lon, time=point.time)
                gpx_segment.points.append(gpx_point)
        return gpx.to_xml()

    def to_life(self):
        """Converts track to LIFE format
        """
        buff = "--%s\n" % self.segments[0].points[0].time.strftime("%Y_%m_%d")
        # buff += "--" + day
        # buff += "UTC+s" # if needed

        def military_time(time):
            """ Converts time to military time

            Args:
                time (:obj:`datetime.datetime`)
            Returns:
                str: Time in the format 1245 (12 hours and 45 minutes)
            """
            return time.strftime("%H%M")

        def stay(buff, start, end, place):
            """ Creates a stay representation

            Args:
                start (:obj:`datetime.datetime` or str)
                end (:obj:`datetime.datetime` or str)
                place (:obj:`Location`)
            Returns:
                str
            """
            if not isinstance(start, str):
                start = military_time(start)
            if not isinstance(end, str):
                end = military_time(end)

            return "%s\n%s-%s: %s" % (buff, start, end, place.label)

        def trip(buff, segment):
            """ Creates a trip representation

            Args:
                buff (str): buffer
                segment (:obj:`Segment`)
            Returns:
                str: buffer and trip representation
            """
            trip = "%s-%s: %s -> %s" % (
                military_time(segment.points[0].time),
                military_time(segment.points[-1].time),
                segment.location_from.label,
                segment.location_to.label
            )

            t_modes = segment.transportation_modes
            if len(t_modes) == 1:
                trip = "%s [%s]" % (trip, t_modes[0]['label'])
            else:
                modes = []
                for mode in t_modes:
                    trip_from = military_time(segment.points[mode['from']].time)
                    trip_to = military_time(segment.points[mode['to']].time)
                    modes.append("    %s-%s: [%s]" % (trip_from, trip_to, mode['label']))
                trip = "%s\n%s" % (trip, "\n".join(modes))

            return "%s\n%s" % (buff, trip)

        last = len(self.segments) - 1
        for i, segment in enumerate(self.segments):
            if i == 0:
                buff = stay(
                    buff,
                    "0000",
                    military_time(segment.points[0].time),
                    segment.location_from
                )
            buff = trip(buff, segment)
            if i is last:
                buff = stay(
                    buff,
                    military_time(segment.points[0].time),
                    "2359",
                    segment.location_to
                )
            else:
                next_seg = self.segments[i+1]
                buff = stay(
                    buff,
                    military_time(segment.points[-1].time),
                    military_time(next_seg.points[0].time),
                    segment.location_from
                )

        return buff

    @staticmethod
    def from_gpx(file_path):
        """ Creates a Track from a GPX file.

        No preprocessing is done.

        Arguments:
            file_path (str): file path and name to the GPX file
        Return:
            :obj:`list` of :obj:`Track`
        """
        gpx = gpxpy.parse(open(file_path, 'r'))
        file_name = basename(file_path)

        tracks = []
        for i, track in enumerate(gpx.tracks):
            segments = []
            for segment in track.segments:
                segments.append(Segment.from_gpx(segment))

            if len(gpx.tracks) > 1:
                name = file_name + "_" + str(i)
            else:
                name = file_name
            tracks.append(Track(name, segments))

        return tracks

    @staticmethod
    def from_json(json):
        """Creates a Track from a JSON file.

        No preprocessing is done.

        Arguments:
            json: map with the keys: name (optional) and segments.
        Return:
            A track instance
        """
        segments = [Segment.from_json(s) for s in json['segments']]
        return Track(json['name'], segments).compute_metrics()



import csv
import sys
from tracktotrip import Track
from os import listdir
from os.path import join
from sklearn import tree
# from sklearn import cross_validation
from sklearn.cross_validation import LabelKFold
from sklearn.cross_validation import train_test_split
"""
Outputs a C4.5 classifier that can be used by TrackToTrip
to identify transportation modes.
"""

def buildDataset(folder, output):
    dataset = {}
    files = listdir(folder)

    l = len(files)
    for i, f in enumerate(files):
        sys.stdout.write('Processing %d of %d: %s       \r' % (i+1, l, f))
        sys.stdout.flush()

        label = f.split('.')[0]
        # content = open(join(folder, f), 'r')
        track = Track.fromGPX(join(folder, f))[0]
        track.preprocess(destructive=False)
        # content.close()

        if not label in dataset:
            dataset[label] = []

        createFeatures(track, dataset, label)

    sys.stdout.write('\nExporting to CSV\n')
    sys.stdout.flush()
    with open(output, 'wb') as csvfile:
        w = csv.writer(csvfile, delimiter=' ')
        for label, features in dataset.iteritems():
            for feature in features:
                row = [label]
                row.extend(feature)
                w.writerow(row)

    print('\nDone')

    return dataset

def createFeatures(track, dataset, label):
    # Assumed that each track has one segment
    segment = track.segments[0]

    # this feature group is composed of (where i is the index):
    # [
    #   vel[i-5], ..., vel[i], ..., vel[i+5],
    #   acc[i-5], ..., acc[i], ..., acc[i+5]
    # ]

    window = 5
    nPoints = len(segment.points) - 1
    for i, point in enumerate(segment.points):
        pointFeatures = []
        for j in range(-window, window + 1):
            index = i + j
            vel = None
            acc = None
            if index > 0 and index <= nPoints:
                vel = segment.points[index].vel
                acc = segment.points[index].acc
            pointFeatures.append(vel)
            pointFeatures.append(acc)
        dataset[label].append(pointFeatures)

def loadDataset(filePath):
    with open(filePath, 'r') as f:
        labels = []
        data = []
        r = csv.reader(f, delimiter=' ')
        for row in r:
            label = row[0]
            d = row[1:]
            labels.append(label)
            data.append(d)

        labelSet = set(labels)
        labelMap = {}
        for i, ls in enumerate(labelSet):
            labelMap[ls] = i

        labels = map(lambda l: labelMap[l], labels)
        return {
                'data': data,
                'target': labels,
                'labels': labelSet
                }

def buildClassifier(dataset):
    clf = tree.DecisionTreeClassifier()
    clf = clf.fit(dataset.data, dataset.target)
    return clf

def plotClassifier(clf, dataset, name='clf'):
    with open("%s.dot" % name, 'w') as f:
        f = tree.export_graphviz(clf, out_file=f,
                         feature_names=dataset.feature_names,
                         class_names=dataset.target_names,
                         filled=True, rounded=True,
                         special_characters=True)

def crossValidate(clf, dataset):
    """Cross validate the classifier

    Uses k-fold method
    """
    X, y = dataset.data, dataset.target
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)
    LabelKFold(dataset.labels, n_folds=4)
    return

def runDataset(dataset):
    clf = tree.DecisionTreeClassifier()
    clf = clf.fit(dataset['data'], dataset['target'])
    print(clf.predict([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]))
# Build dataset
# buildDataset('./dataset', './dataset.csv')
runDataset(loadDataset('./dataset.csv'))



# SMOOTH_STRATEGY = 'inverse'
# SMOOTH_NOISE = 1.0
#
# SEGMENT_EPS = 0.15
# SEGMENT_MIN_SAMPLES = 80
# SEGMENT_MIN_TIME = 60
#
# # SIMPLIFY_MAX_TIME = 5
# # SIMPLIFY_MAX_DISTANCE = 0.01
# SIMPLIFY_DISTANCE_THRESHOLD = 30
# SIMPLIFY_EPS = 0.0001
#
# TM_REMOVE_STOPS = False
# TM_DT_THRESHOLD = 10
#
# LOCATIONS_LIMIT = 5
# LOCATION_MAX_DISTANCE = 20 # Distance in meters
#
# PREPROCESS_MAX_ACC = 100000.0



"""
Spatio-temporal segmentation of points
"""
import numpy as np
from sklearn.cluster import DBSCAN
from sklearn.preprocessing import StandardScaler

def correct_segmentation(segments, clusters, min_time):
    """ Corrects the predicted segmentation

    This process prevents over segmentation

    Args:
        segments (:obj:`list` of :obj:`list` of :obj:`Point`):
            segments to correct
        min_time (int): minimum required time for segmentation
    """
    segments = [points for points in segments if len(points) > 1]

    result_segments = []
    prev_segment = None
    for i, segment in enumerate(segments):
        cluster = clusters[i]
        if prev_segment is None:
            prev_segment = segment
        else:
            cluster_dt = 0
            if len(cluster) > 0:
                cluster_dt = abs(cluster[0].time_difference(cluster[-1]))
            print(i, cluster_dt, cluster_dt <= min_time)
            if cluster_dt <= min_time:
                prev_segment.extend(segment)
            else:
                prev_segment.append(segment[0])
                result_segments.append(prev_segment)
                prev_segment = segment
    if prev_segment is not None:
        result_segments.append(prev_segment)

    return result_segments

def spatiotemporal_segmentation(points, eps, min_time):
    """ Splits a set of points into multiple sets of points based on
        spatio-temporal stays

    DBSCAN is used to predict possible segmentations,
        furthermore we check to see if each clusters is big enough in
        time (>=min_time). If that's the case than the segmentation is
        considered valid.

    When segmenting, the last point of the ith segment will be the same
        of the (i-1)th segment.

    Segments are identified through clusters.
    The last point of a clusters, that comes after a sub-segment A, will
        be present on the sub-segment A.

    Args:
        points (:obj:`list` of :obj:`Point`): segment's points
        eps (float): Epsilon to feed to the DBSCAN algorithm.
            Maximum distance between two samples, to be considered in
            the same cluster.
        min_time (float): Minimum time of a stay
    Returns:
        :obj:`list` of :obj:`list` of :obj:`Point`: Initial set of
            points in different segments
    """
    # min time / sample rate
    min_samples = min_time / np.average([point.dt for point in points])

    data = [point.gen3arr() for point in points]
    data = StandardScaler().fit_transform(data)
    db_cluster = DBSCAN(eps=eps, min_samples=min_samples).fit(data)
    labels = db_cluster.labels_

    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)

    segments = [[] for _ in range(n_clusters_+1)]
    clusters = [[] for _ in range(n_clusters_+1)]
    current_segment = 0

    # split segments identified with dbscan
    for i, label in enumerate(labels):
        if label != -1 and label + 1 != current_segment:
            current_segment = label + 1
        point = points[i]
        segments[current_segment].append(point)
        if label == -1:
            pass
        else:
            clusters[label + 1].append(point)

    return correct_segmentation(segments, clusters, min_time)



"""
Compression algorithms

There are two distinct types:
    - topology based, such as douglas ramer peucker
    - time based, which are represented by td_sp, td_tr and the combination of both, spt
"""
import sys
from math import sqrt
from .point import Point

# All of the compression methods require recursion.
# Tracks with a huge number of points the default recursion limit (1000) could be a problem
sys.setrecursionlimit(10000)

I_3600 = 1 / 3600.0

def loc_dist(end, start):
    """ Spatial distance between two points (end-start)

    Args:
        start (:obj:`Point`)
        end (:obj:`Point`)
    Returns:
        float, distance in m
    """
    return end.distance(start)

def time_dist(end, start):
    """ Temporal distance between two points (end-start)

    Args:
        start (:obj:`Point`)
        end (:obj:`Point`)
    Returns:
        float, time difference in seconds
    """
    return end.time_difference(start)

def distance(p_a, p_b):
    """ Euclidean distance, between two points

    Args:
        p_a (:obj:`Point`)
        p_b (:obj:`Point`)
    Returns:
        float: distance, in degrees
    """
    return sqrt((p_a.lat - p_b.lat) ** 2 + (p_a.lon - p_b.lon) ** 2)

def point_line_distance(point, start, end):
    """ Distance from a point to a line, formed by two points

    Args:
        point (:obj:`Point`)
        start (:obj:`Point`): line point
        end (:obj:`Point`): line point
    Returns:
        float: distance to line, in degrees
    """
    if start == end:
        return distance(point, start)
    else:
        un_dist = abs(
            (end.lat-start.lat)*(start.lon-point.lon) - (start.lat-point.lat)*(end.lon-start.lon)
        )
        n_dist = sqrt(
            (end.lat-start.lat)**2 + (end.lon-start.lon)**2
        )
        return un_dist / n_dist

def drp(points, epsilon):
    """ Douglas ramer peucker

    Based on https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm

    Args:
        points (:obj:`list` of :obj:`Point`)
        epsilon (float): drp threshold
    Returns:
        :obj:`list` of :obj:`Point`
    """
    dmax = 0.0
    index = 0

    for i in range(1, len(points)-1):
        dist = point_line_distance(points[i], points[0], points[-1])
        if dist > dmax:
            index = i
            dmax = dist

    if dmax > epsilon:
        return drp(points[:index+1], epsilon)[:-1] + drp(points[index:], epsilon)
    else:
        return [points[0], points[-1]]

def td_sp(points, speed_threshold):
    """ Top-Down Speed-Based Trajectory Compression Algorithm

    Detailed in https://www.itc.nl/library/Papers_2003/peer_ref_conf/meratnia_new.pdf

    Args:
        points (:obj:`list` of :obj:`Point`): trajectory or part of it
        speed_threshold (float): max speed error, in km/h
    Returns:
        :obj:`list` of :obj:`Point`, compressed trajectory
    """
    if len(points) <= 2:
        return points
    else:
        max_speed_threshold = 0
        found_index = 0
        for i in range(1, len(points)-1):
            dt1 = time_dist(points[i], points[i-1])
            if dt1 == 0:
                dt1 = 0.000000001
            vim = loc_dist(points[i], points[i-1]) / dt1
            dt2 = time_dist(points[i+1], points[i])
            if dt2 == 0:
                dt2 = 0.000000001
            vi_ = loc_dist(points[i+1], points[i]) / dt2
            if abs(vi_ - vim) > max_speed_threshold:
                max_speed_threshold = abs(vi_ - vim)
                found_index = i
        if max_speed_threshold > speed_threshold:
            one = td_sp(points[:found_index], speed_threshold)
            two = td_sp(points[found_index:], speed_threshold)
            one.extend(two)
            return one
        else:
            return [points[0], points[-1]]

def td_tr(points, dist_threshold):
    """ Top-Down Time-Ratio Trajectory Compression Algorithm

    Detailed in https://www.itc.nl/library/Papers_2003/peer_ref_conf/meratnia_new.pdf

    Args:
        points (:obj:`list` of :obj:`Point`): trajectory or part of it
        dist_threshold (float): max distance error, in meters
    Returns:
        :obj:`list` of :obj:`Point`, compressed trajectory
    """
    if len(points) <= 2:
        return points
    else:
        max_dist_threshold = 0
        found_index = 0
        delta_e = time_dist(points[-1], points[0]) * I_3600
        d_lat = points[-1].lat - points[0].lat
        d_lon = points[-1].lon - points[0].lon

        for i in range(1, len(points)-1):
            delta_i = time_dist(points[i], points[0]) * I_3600

            di_de = delta_i / delta_e
            point = Point(
                points[0].lat + d_lat * di_de,
                points[0].lon + d_lon * di_de,
                None
            )

            dist = loc_dist(points[i], point)
            if dist > max_dist_threshold:
                max_dist_threshold = dist
                found_index = i

        if max_dist_threshold > dist_threshold:
            one = td_tr(points[:found_index], dist_threshold)
            two = td_tr(points[found_index:], dist_threshold)
            one.extend(two)
            return one
        else:
            return [points[0], points[-1]]

def spt(points, max_dist_error, max_speed_error):
    """ A combination of both `td_sp` and `td_tr`

    Detailed in,
        Spatiotemporal Compression Techniques for Moving Point Objects,
        Nirvana Meratnia and Rolf A. de By, 2004,
        in Advances in Database Technology - EDBT 2004: 9th
        International Conference on Extending Database Technology,
        Heraklion, Crete, Greece, March 14-18, 2004

    Args:
        points (:obj:`list` of :obj:`Point`)
        max_dist_error (float): max distance error, in meters
        max_speed_error (float): max speed error, in km/h
    Returns:
        :obj:`list` of :obj:`Point`
    """
    if len(points) <= 2:
        return points
    else:
        is_error = False
        e = 1
        while e < len(points) and not is_error:
            i = 1
            while i < e and not is_error:
                delta_e = time_dist(points[e], points[0]) * I_3600
                delta_i = time_dist(points[i], points[0]) * I_3600

                di_de = delta_i / delta_e
                d_lat = points[e].lat - points[0].lat
                d_lon = points[e].lon - points[0].lon
                point = Point(
                    points[0].lat + d_lat * di_de,
                    points[0].lon + d_lon * di_de,
                    None
                )

                dt1 = time_dist(points[i], points[i-1])
                if dt1 == 0:
                    dt1 = 0.000000001
                dt2 = time_dist(points[i+1], points[i])
                if dt2 == 0:
                    dt2 = 0.000000001

                v_i_1 = loc_dist(points[i], points[i-1]) / dt1
                v_i = loc_dist(points[i+1], points[i]) / dt2

                if loc_dist(points[i], point) > max_dist_error or abs(v_i - v_i_1) > max_speed_error:
                    is_error = True
                else:
                    i = i + 1
            if is_error:
                return [points[0]] + spt(points[i:len(points)], max_dist_error, max_speed_error)
            e = e + 1
        if not is_error:
            return [points[0], points[len(points)-1]]



"""
Compatibility Support for Python 2.7 and earlier
"""

import sys


def get_all_headers(message, key):
    """
    Given an HTTPMessage, return all headers matching a given key.
    """
    return message.get_all(key)


if sys.version_info < (3,):
    def get_all_headers(message, key):
        return message.getheaders(key)



"""
MIT License
"""

import pickle
from sklearn import preprocessing
from sklearn.linear_model import SGDClassifier

# from random import shuffle
# import numpy as np
# import matplotlib.pyplot as plt


class Classifier(object):
    """ Transportation mode classifier

    This is a wrapper around sklearn SGDClassifier, to simplify it's usage

    Attributes:
        clf (:obj:`SGDClassifier`): Classifier being used
        labels (:obj:`LabelEncoder`): Label encoder, includes all the labels
        feature_length (int): Length of each feature. <0 if it hasn't learned any
    """
    def __init__(self):
        self.clf = SGDClassifier(loss="log", penalty="l1", shuffle=True)
        self.labels = preprocessing.LabelEncoder()
        self.feature_length = -1

    def __learn_labels(self, labels):
        """ Learns new labels, this method is intended for internal use

        Args:
            labels (:obj:`list` of :obj:`str`): Labels to learn
        """
        if self.feature_length > 0:
            result = list(self.labels.classes_)
        else:
            result = []

        for label in labels:
            result.append(label)
        self.labels.fit(result)

    def learn(self, features, labels):
        """ Fits the classifier

        If it's state is empty, the classifier is fitted, if not
        the classifier is partially fitted.
        See sklearn's SGDClassifier fit and partial_fit methods.

        Args:
            features (:obj:`list` of :obj:`list` of :obj:`float`)
            labels (:obj:`list` of :obj:`str`): Labels for each set of features.
                New features are learnt.
        """
        self.__learn_labels(labels)
        labels = self.labels.transform(labels)
        if self.feature_length > 0:
            # FIXME? check docs, may need to pass class=[...]
            self.clf = self.clf.partial_fit(features, labels)
        else:
            self.clf = self.clf.fit(features, labels)
            self.feature_length = len(features[0])

    def predict(self, features, verbose=False):
        """ Probability estimates of each feature

        See sklearn's SGDClassifier predict and predict_proba methods.

        Args:
            features (:obj:`list` of :obj:`list` of :obj:`float`)
            verbose: Boolean, optional. If true returns an array where each
                element is a dictionary, where keys are labels and values are
                the respective probabilities. Defaults to False.
        Returns:
            Array of array of numbers, or array of dictionaries if verbose i
            True
        """
        probs = self.clf.predict_proba(features)
        if verbose:
            labels = self.labels.classes_
            res = []
            for prob in probs:
                vals = {}
                for i, val in enumerate(prob):
                    label = labels[i]
                    vals[label] = val
                res.append(vals)
            return res
        else:
            return probs

    def save_to_file(self, filename):
        """ Saves this instance to a file

        Args:
            filename (str)
        """
        pickle.dump(self, filename)

    @staticmethod
    def load_from_file(filename):
        """ Loads a classifier instance from a file

        Args:
            filename (str)
        """
        return pickle.load(filename)

    # def plot(self, colors='rbg', max=10):
    #     result = []
    #     X = np.arange(0, max, 0.1)
    #     for i in X:
    #         [[a, b, c]] = self.predict([[i]])
    #         result.append((a, b, c))
    #     for y in range(3):
    #         plt.plot(X, map(lambda i: i[y], result), '-' + colors[y])


# data = {
#     'stop': [[0], [0.01], [0.05]],
#     'walk': [[0.5], [3], [4], [5]],
#     'car': [[7], [8], [9], [10], [20]]
#     }

# c = Classifier()
#
# data_labels = []
# data_values = []
# items = data.items()
# print(items)
# items = sorted(items, key=lambda i: i[0])
# for label, values in items:
#     for value in values:
#         data_labels.append(label)
#         data_values.append(value)
#
# print(data_labels, data_values)
# c.learn(data_values, data_labels)
# print(c.predict([[10]], verbose=True))
# c.learn([[1]], ['kid'])
# print(c.predict([[10]], verbose=True))
# c.plot(max=30)


# plot_clf(clf)
# plt.show()



from tracktotrip.segmentSimilarity import line, angle_similarity, distanceToLine, distance_similarity, normalize, lineDistance_similarity, line_similarity, sortSegmentPoints, segment_similarity
import unittest
import math
# import matplotlib.pyplot as plt
from tracktotrip import Segment, Track, Point, kalman

class TestSimilarity(unittest.TestCase):
    def test_angle__similarity(self):
        l1p1 = [10, 10]
        l1p2 = [10, 12]
        l1 = line(l1p1, l1p2)

        l2p1 = [11, 10]
        l2p2 = [10, 10]
        l2 = line(l2p1, l2p2)

        self.assertEqual(angle_similarity(normalize(l1), normalize(l2)), 0)
        self.assertEqual(angle_similarity(normalize([0, 1]), normalize([1, 1])), angle_similarity(normalize([1, 0]), normalize([1, 1])))
        self.assertEqual(angle_similarity(normalize([0, 1]), normalize([0, -1])), -1)

    def test_distance(self):
        sqrt2 = math.sqrt(2)
        self.assertEqual(distanceToLine([0,0], [4,2], [0,0]), 0)
        self.assertEqual(distanceToLine([0,0], [4,2], [4,2]), 0)
        self.assertEqual(distanceToLine([0,0], [4,2], [2,1]), 0)
        self.assertEqual(distanceToLine([0,0], [2,2], [0,2]), sqrt2)
        self.assertEqual(distanceToLine([0,0], [2,2], [3,3]), sqrt2)

    def test_distance__similarity(self):
        self.assertEqual(distance_similarity([0,0], [4,2], [0,0], T=1), 1)
        self.assertEqual(distance_similarity([0,0], [4,2], [4,2], T=1), 1)
        self.assertEqual(distance_similarity([0,0], [4,2], [2,1], T=1), 1)

        self.assertEqual(distance_similarity([0,0], [2,0], [0,1], T=1), 0)
        self.assertEqual(distance_similarity([0,0], [2,0], [0,0.5], T=1), 0.5)
        self.assertEqual(distance_similarity([0,0], [2,0], [0,-1], T=1), 0)

        self.assertEqual(distance_similarity([0,0], [2,0], [0,0.5], T=10.0), 0.95)

    def test_line_distance__similarity(self):
        self.assertEqual(lineDistance_similarity([0, 0], [5, 2], [0, 0], [5, 2], T=1), 1)
        self.assertEqual(lineDistance_similarity([0, 0], [5, 2], [5, 2], [0, 0], T=1), 1)
        self.assertTrue(lineDistance_similarity([0, 0], [5, 2], [0, 0], [5, 1.9], T=1) > 0.95)
        self.assertTrue(lineDistance_similarity([0, 0], [5, 2], [0, 0], [5, 2.1], T=1) )
        self.assertTrue(lineDistance_similarity([0, 0], [1, 1], [1, 0], [0, 1], T=1) < 0.3)

    def test_line__similarity(self):
        self.assertTrue(isclose(line_similarity([0, 0], [5, 2], [0, 0], [5, 2]), 1))
        self.assertTrue(isclose(line_similarity([0, 0], [5, 2], [5, 2], [0, 0]), 1))
        self.assertTrue(isclose(line_similarity([0, 0], [5, 2], [0, 0], [5, 1.9]), 0.9534, err=0.01))
        self.assertTrue(isclose(line_similarity([0, 0], [5, 2], [0, 0], [5, 2.1]), 0.9498, err=0.01))
        self.assertTrue(isclose(line_similarity([0, 0], [1, 1], [1, 0], [0, 1]), 0))

    def test_segment__similarity(self):
        Aps = [[0.5, 0.5], [1, 1.5], [2, 2.5], [3.5, 3.5], [5.2, 4.5], [7.5, 6.5], [7.9, 8]]
        Bps = [[0.6, 0.5], [1.05, 1.45], [2.1, 2.4], [2.8, 4], [3.5, 5.5], [5, 5.7], [7.8, 5.7], [8.1, 6.5], [8.1, 8]]

        A = Segment(map(lambda p: Point(None, p[0], p[1], None), Aps))
        B = Segment(map(lambda p: Point(None, p[0], p[1], None), Bps))
        similarity, parts = segment_similarity(A, B)

        self.assertTrue(similarity > 0.35)
        self.assertTrue(parts[3] > parts[4])
        self.assertTrue(parts[4] < parts[5])
        # mid = sortSegmentPoints(Aps, Bps)

        # def plot(data, more=""):
            # plt.plot(map(lambda a: a[0], data), map(lambda a: a[1], data), more)
        # plot(Aps)
        # plot(Bps)

        # plot(mid)
        # kalm_res = kalman.kalman(mid)
        # plot(kalm_res, "ko-")

        # plt.show()



def isclose(a, b, err=1e-09):
    return abs(a-b) <= err

if __name__ == '__main__':
    unittest.main()

def testWithTracks():
    track1 = Track.fromGPX('/Users/ruipgil/tracks_bckp/2015-07-23_1.gpx')[0]
    track2 = Track.fromGPX('/Users/ruipgil/tracks_bckp/2015-07-23_3.gpx')[0]
    track1.preprocess()
    track2.preprocess()

    track1.toTrip()
    track2.toTrip()

    t1p = map(lambda p: p.gen2arr(), track1.segments[0].points)
    t2p = map(lambda p: p.gen2arr(), track2.segments[0].points)

    globalprox, part = segment_similarity(track1.segments[0], track2.segments[0])
    print(globalprox)

    # res = kalman(init)

    plt.axis('equal')
    def plot(data, more=""):
        plt.plot(map(lambda a: a[0], data), map(lambda a: a[1], data), more)
        # plt.plot(data[0][0], data[0][1], more+'o')
        # plt.plot(data[-1][0], data[-1][1], more+'o')
    plot(t1p)
    plot(t2p, 'k-')

    overlaps = []
    group = []
    for i, a in enumerate(part):
        if a > 0:
            group.append(t2p[i])
        elif len(group) > 0:
            overlaps.append(group)
            for j in group:
                plot(group)
            group = []
    plot(group)
    overlaps.append(group)

    print(len(overlaps))
    plt.show()

testWithTracks()



from tracktotrip.classifier import Classifier
import pickle

features = pickle.load(open('geolife.features', 'r'))
labels = pickle.load(open('geolife.labels', 'r'))

print('Classifier built')
clf = Classifier()
clf.learn(features, labels)
clf.save_to_file(open('classifier.data', 'w'))

# print(clf.predict([0, 0.7, 0, 0.2, ]))
# print(clf.predict([1, 0.7]))

# from sklearn.cross_validation import KFold
#
# kf = KFold(4, n_folds=2)
# for train_index, test_index in kf:
#     X_train, X_test = X[train_index], X[test_index]
#     y_train, y_test = y[train_index], y[test_index]
#
# scores = cross_validation.cross_val_score(
#     clf,
#     iris.data,
#     iris.target,
#     cv=5,
#     scoring='f1_weighted'
# )




"""
Kalman filter
See https://github.com/open-city/ikalman
"""
import ikalman

def kalman_filter(points, noise):
    """ Smooths points with kalman filter

    See https://github.com/open-city/ikalman

    Args:
        points (:obj:`list` of :obj:`Point`): points to smooth
        noise (float): expected noise
    """
    kalman = ikalman.filter(noise)
    for point in points:
        kalman.update_velocity2d(point.lat, point.lon, point.dt)
        (lat, lon) = kalman.get_lat_long()
        point.lat = lat
        point.lon = lon
    return points



"""
Location class and methods
"""
from math import sqrt
import requests
import numpy as np
from sklearn.cluster import DBSCAN
from .utils import estimate_meters_to_deg


GOOGLE_PLACES_URL = 'https://maps.googleapis.com/maps/api/place/nearbysearch' \
    '/json?location=%s,%s&radius=%s&key=%s'

def compute_centroid(points):
    """ Computes the centroid of set of points

    Args:
        points (:obj:`list` of [float, float])
    Returns:
        [float, float]: Latitude and longitude of the centroid
    """
    lats = [p[0] for p in points]
    lons = [p[1] for p in points]
    return [np.mean(lats), np.mean(lons)]

def update_location_centroid(point, cluster, max_distance, min_samples):
    """ Updates the centroid of a location cluster with another point

    Args:
        point (:obj:`Point`): Point to add to the cluster
        cluster (:obj:`list` of :obj:`Point`): Location cluster
        max_distance (float): Max neighbour distance
        min_samples (int): Minimum number of samples
    Returns:
        ([float, float], :obj:`list` of :obj:`Point`): Tuple with the location centroid
            and new point cluster (given cluster + given point)
    """
    points = [p.gen2arr() for p in cluster]
    points.append(point.gen2arr())

    # Estimates the epsilon
    eps = estimate_meters_to_deg(sqrt(max_distance), precision=5)

    p_cluster = DBSCAN(eps=eps, min_samples=min_samples)
    p_cluster.fit(points)

    clusters = {}
    for i, label in enumerate(p_cluster.labels_):
        if label in clusters.keys():
            clusters[label].append(points[i])
        else:
            clusters[label] = [points[i]]

    centroids = []
    biggest_centroid_l = -float("inf")
    biggest_centroid = None

    for label, cluster in clusters.items():
        centroid = compute_centroid(cluster)
        centroids.append(centroid)

        if label >= 0 and len(cluster) >= biggest_centroid_l:
            biggest_centroid_l = len(cluster)
            biggest_centroid = centroid

    if biggest_centroid is None:
        biggest_centroid = compute_centroid(points)

    return biggest_centroid, points


def query_google(point, max_distance, key):
    """ Queries google maps API for a location

    Args:
        point (:obj:`Point`): Point location to query
        max_distance (float): Search radius, in meters
        key (str): Valid google maps api key
    Returns:
        :obj:`list` of :obj:`dict`: List of locations with the following format:
            {
                'label': 'Coffee house',
                'types': 'Commerce',
                'suggestion_type': 'GOOGLE'
            }
    """
    if not key:
        return []

    req = requests.get(GOOGLE_PLACES_URL % (
        point.lat,
        point.lon,
        max_distance,
        key
    ))

    if req.status_code != 200:
        return []
    response = req.json()
    results = response['results']
    # l = len(results)
    final_results = []
    for local in results:
        final_results.append({
            'label': local['name'],
            # 'rank': (l-i)/float(l),
            'types': local['types'],
            'suggestion_type': 'GOOGLE'
            })
    return final_results

def infer_location(point, location_query, max_distance, google_key, limit):
    """ Infers the semantic location of a (point) place.

    Args:
        points (:obj:`Point`): Point location to infer
        location_query: Function with signature, (:obj:`Point`, int) -> (str, :obj:`Point`, ...)
        max_distance (float): Max distance to a position, in meters
        google_key (str): Valid google maps api key
        limit (int): Results limit
    Returns:
        :obj:`Location`: with top match, and alternatives
    """
    locations = []

    if location_query is not None:
        queried_locations = location_query(point, max_distance)
        for (label, centroid, _) in queried_locations:
            locations.append({
                'label': label,
                'distance': centroid.distance(point),
                'cantroid': centroid,
                'suggestion_type': 'KB'
                })
        locations = sorted(locations, key=lambda d: d['distance'])

    if len(locations) <= limit:
        google_locs = query_google(point, max_distance, google_key)
        locations.extend(google_locs)

    locations = locations[:limit]

    return Location(locations[0]['label'], point, locations)


class Location(object):
    """ Location representation

    Params:
        label (str): Location name
        centroid (:obj:`Point`): Location position
        other (:obj:`list` of :obj:`dict`): Other possible locations. Includes the current label
    """
    def __init__(self, label, position, other):
        self.label = label
        self.centroid = position
        self.other = other

    def distance(self, position):
        """ Computes the distance between centroid and another point

        Args:
            position (:obj:`Point`)
        Returns:
            float: distance, in meters
        """
        return self.centroid.distance(position)

    def to_json(self):
        """ Converts to a json representation

        Returns:
            :obj:`dict`
        """
        return {
            'label': self.label,
            'position': self.centroid.to_json(),
            'other': self.other
        }

    @staticmethod
    def from_json(json):
        """ Converts from a json representation

        Returns:
            :obj:`Location`
        """
        return Location(json['label'], json['position'], [])



"""
Time-distance compression algorithms
"""
from .point import Point

def loc_dist(end, start):
    """ Spatial distance between two points (end-start)

    Args:
        start (:obj:`Point`)
        end (:obj:`Point`)
    Returns:
        float, distance in m
    """
    return end.distance(start)

def time_dist(end, start):
    """ Temporal distance between two points (end-start)

    Args:
        start (:obj:`Point`)
        end (:obj:`Point`)
    Returns:
        float, time difference in seconds
    """
    return end.time_difference(start)

def td_sp(points, speed_threshold):
    """ Top-Down Speed-Based Trajectory Compression Algorithm

    Detailed in https://www.itc.nl/library/Papers_2003/peer_ref_conf/meratnia_new.pdf

    Args:
        points (:obj:`list` of :obj:`Point`): trajectory or part of it
        speed_threshold (float): speed threshold
    Returns:
        :obj:`list` of :obj:`Point`, compressed trajectory
    """
    if len(points) <= 2:
        return points
    else:
        max_speed_threshold = 0
        found_index = 0
        for i in range(1, len(points)-1):
            dt1 = time_dist(points[i], points[i-1])
            if dt1 == 0:
                dt1 = 0.000000001
            vim = loc_dist(points[i], points[i-1]) / dt1
            dt2 = time_dist(points[i+1], points[i])
            if dt2 == 0:
                dt2 = 0.000000001
            vi_ = loc_dist(points[i+1], points[i]) / dt2
            if abs(vi_ - vim) > max_speed_threshold:
                max_speed_threshold = abs(vi_ - vim)
                found_index = i
        if max_speed_threshold > speed_threshold:
            one = td_sp(points[:found_index], speed_threshold)
            two = td_sp(points[found_index:], speed_threshold)
            one.extend(two)
            return one
        else:
            return [points[0], points[-1]]

def td_tr(points, dist_threshold):
    """ Top-Down Time-Ratio Trajectory Compression Algorithm

    Detailed in https://www.itc.nl/library/Papers_2003/peer_ref_conf/meratnia_new.pdf

    Args:
        points (:obj:`list` of :obj:`Point`): trajectory or part of it
        dist_threshold (float): distance threshold
    Returns:
        :obj:`list` of :obj:`Point`, compressed trajectory
    """
    if len(points) <= 2:
        return points
    else:
        max_dist_threshold = 0
        found_index = 0
        delta_e = time_dist(points[-1], points[0]) / 3600
        d_lat = points[-1].lat - points[0].lat
        d_lon = points[-1].lon - points[0].lon

        for i in range(1, len(points)-1):
            delta_i = time_dist(points[i], points[0]) / 3600

            di_de = delta_i / delta_e
            point = Point(
                points[0].lat + d_lat * di_de,
                points[0].lon + d_lon * di_de,
                None
            )

            dist = loc_dist(points[i], point)
            if dist > max_dist_threshold:
                max_dist_threshold = dist
                found_index = i

        if max_dist_threshold > dist_threshold:
            one = td_tr(points[:found_index], dist_threshold)
            two = td_tr(points[found_index:], dist_threshold)
            one.extend(two)
            return one
        else:
            return [points[0], points[-1]]



"""
Setup script
"""
import os
from distutils.core import setup

def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()

setup(
    name='tracktotrip',
    packages=['tracktotrip'],
    version='0.3',
    description='Track processing library',
    author='Rui Gil',
    author_email='ruipgil@gmail.com',
    url='https://github.com/ruipgil/TrackToTrip',
    download_url='https://github.com/ruipgil/TrackToTrip/archive/master.zip',
    keywords=['track', 'trip', 'GPS', 'GPX'],
    classifiers=[],
    install_requires=read('requirements.txt').split('\n')
)



"""
Segment preprocessing
"""
# import datetime
# from .point import Point
#
# MIN_TIME = datetime.datetime(1999, 1, 1)
#
# def preprocess_segment(points, max_acc, destructive=True):
#     """ Calculates the metrics of points and unrealistic ones
#
#     Uses their ordering to calculate velocities and spatio-temporal distances
#     If destructive it removes points with accelarations higher than max_acc
#         and that occured before MIN_TIME
#
#     Args:
#         points (:obj:`list` of :obj:`Point`)
#         max_acc (float): Maximum acceleration threshold
#         destructive (bool, optional): True to allow the removal of elements
#             from the initial array. Default is True
#     Returns:
#         A tuple with the resulting array of points and the
#         the points dropped
#     """
#     result = []
#     last = None
#     for point in points:
#         if not (destructive and point.time < MIN_TIME):
#             current = Point(point.lat, point.lon, point.time)
#
#             if last is not None:
#                 current.compute_metrics(last)
#
#             if destructive and abs(current.acc) > max_acc:
#                 current = Point(last.lat, last.lon, point.time)
#                 current.compute_metrics(last)
#
#             result.append(current)
#             last = current
#
#     return result



import gpxpy
import unittest
from tracktotrip import Point, Segment
from datetime import datetime, timedelta

def loadFile(name):
    return gpxpy.parse(open(name, 'r'))

class TestSegment(unittest.TestCase):
    def test_creation(self):
        s1 = Segment()
        s2 = Segment([Point(None, 0, 0, None)])

        self.assertEqual(len(s1.points), 0)
        self.assertEqual(len(s1.transportation_modes), 0)
        self.assertEqual(s1.location_from, None)
        self.assertEqual(s1.location_to, None)

        self.assertEqual(len(s2.points), 1)
        self.assertEqual(len(s2.transportation_modes), 0)
        self.assertEqual(s2.location_from, None)
        self.assertEqual(s2.location_to, None)

    def test_to_json(self):
        time = datetime.now()
        dt = timedelta(1000)

        s = Segment([Point(None, 0, 0, time), Point(None, 0, 1, time + dt), Point(None, 0, 2, time + dt)])

        json = s.toJSON()
        self.assertTrue('points' in json)
        self.assertTrue('transportationModes' in json)
        self.assertTrue('locationFrom' in json)
        self.assertTrue('locationTo' in json)

if __name__ == '__main__':
    unittest.main()



"""
Point segment module
"""
from copy import deepcopy

import numpy as np

from .drp import drp
from .point import Point
from .smooth import with_extrapolation, with_inverse, INVERSE_STRATEGY, EXTRAPOLATE_STRATEGY
from .td_compression import td_sp
from .preprocess import preprocess_segment
from .location import infer_location
from .transportation_mode import speed_clustering
from .spatiotemporal_segmentation import spatiotemporal_segmentation
from .similarity import sort_segment_points

# from .noise_detection import remove_noise

class Segment(object):
    """Holds the points and semantic information about them

    Attributes:
        points (:obj:`list` of :obj:`Point`): points of the segment
        #TODO
        transportation_modes: array of transportation modes of the segment
            Each transportation mode represents a span of points
            Each span is a map in the following format:
                label: string with the type of transportation mode
                from: start of the span
                to: end of the span
        locationFrom: TrackToTrip.Location or None, the semantic location of
            the start of the segment
        locationTo: TrackToTrip.Location or None, the semantic location of
            the end of the segment
    """

    def __init__(self, points):
        self.points = points
        self.transportation_modes = []
        self.location_from = None
        self.location_to = None

    def bounds(self, lower_index=0, upper_index=-1):
        """ Computes the bounds of the segment, or part of it

        Args:
            lower_index (int, optional): Start index. Defaults to 0
            upper_index (int, optional): End index. Defaults to 0
        Returns:
            :obj:`tuple` of :obj:`float`: Bounds of the (sub)segment, such that
                (min_lat, min_lon, max_lat, max_lon)
        """
        points = self.points[lower_index:upper_index]

        min_lat = float("inf")
        min_lon = float("inf")
        max_lat = -float("inf")
        max_lon = -float("inf")

        for point in points:
            min_lat = min(min_lat, point.lat)
            min_lon = min(min_lon, point.lon)
            max_lat = max(max_lat, point.lat)
            max_lon = max(max_lon, point.lon)

        return (min_lat, min_lon, max_lat, max_lon)

    # TODO
    # def removeNoise(self, var=2):
    #     """In-place removal of noise points
    #
    #     Applies removeNoise function to points
    #
    #     Returns:
    #         This segment
    #     """
    #     self.points = remove_noise(self.points, var=var)
    #     return self

    def smooth(self, noise, strategy=INVERSE_STRATEGY):
        """ In-place smoothing

        See smooth_segment function

        Args:
            noise (float): Noise expected
            strategy (int): Strategy to use. Either smooth.INVERSE_STRATEGY
                or smooth.EXTRAPOLATE_STRATEGY
        Returns:
            :obj:`Segment`
        """
        if strategy is INVERSE_STRATEGY:
            self.points = with_inverse(self.points, noise)
        elif strategy is EXTRAPOLATE_STRATEGY:
            self.points = with_extrapolation(self.points, noise, 30)
        return self

    def segment(self, eps, min_time):
        """Spatio-temporal segmentation

        See spatiotemporal_segmentation function

        Args:
            eps (float): Maximum distance between two samples
            min_time (float): Minimum time between to segment
        Returns:
            :obj:`list` of :obj:`Point`
        """
        return spatiotemporal_segmentation(self.points, eps, min_time)

    def simplify(self, eps, dist_threshold, topology_only=False):
        """ In-place segment simplification

        See drp and td_sp functions

        Args:
            topology_only (bool, optional): True to only keep topology, not considering
                times when simplifying. Defaults to False.
            dist_threshold (float, optional): Distance threshold for the td_sp function
            eps (float, optional): Distance thresgold for the drp function
        Returns:
            :obj:`Segment`
        """
        if topology_only:
            self.points = drp(self.points, eps)
        else:
            self.points = td_sp(self.points, dist_threshold)
        return self

    def compute_metrics(self):
        """ Computes metrics for each point

        Returns:
            :obj:`Segment`: self
        """
        prev = None
        for point in self.points:
            if prev is None:
                prev = point
            else:
                point.compute_metrics(prev)
                prev = point
        return self

    def preprocess(self, max_acc, destructive=True):
        """In-place segment preprocessing

        See preprocess_segment function

        Args:
            max_acc (float): Max acceleration threshold.
            destructive (bool, optional): Remove points. Defauts to True
        Returns:
            :obj:`Segment`: self
        """
        points = preprocess_segment(self.points, max_acc, destructive)
        self.points = points
        return self

    def infer_location(self, location_query, max_distance, google_key, limit):
        """In-place location inferring

        See infer_location function

        Args:
        Returns:
            :obj:`Segment`: self
        """

        self.location_from = infer_location(
            self.points[0],
            location_query,
            max_distance=max_distance,
            google_key=google_key,
            limit=limit
        )
        self.location_to = infer_location(
            self.points[-1],
            location_query,
            max_distance=max_distance,
            google_key=google_key,
            limit=limit
        )

        return self

    def infer_transportation_mode(self, clf, min_time):
        """In-place transportation mode inferring

        See infer_transportation_mode function

        Args:
        Returns:
            :obj:`Segment`: self
        """
        self.transportation_modes = speed_clustering(clf, self.points, min_time)
        return self

    def merge_and_fit(self, segment):
        """ Merges another segment with this one, ordering the points based on a
            distance heuristic

        Args:
            segment (:obj:`Segment`): Segment to merge with
        Returns:
            :obj:`Segment`: self
        """
        self.points = sort_segment_points(self.points, segment.points)
        return self

    def closest_point_to(self, point, thr=20.0):
        """ Finds the closest point in the segment to a given point

        Args:
            point (:obj:`Point`)
            thr (float, optional): Distance threshold, in meters, to be considered
                the same point. Defaults to 20.0
        Returns:
            int: Index of the point. -1 if doesn't exist
        """
        distances = [p.distance(point) for p in self.points]
        min_index = np.argmin(distances)

        if distances[min_index] > thr:
            return -1
        else:
            return min_index

    def slice(self, start, end):
        """ Creates a copy of the current segment between indexes. If end > start,
            points are reverted

        Args:
            start (int): Start index
            end (int): End index
        Returns:
            :obj:`Segment`
        """

        reverse = False
        if start > end:
            temp = start
            start = end
            end = temp
            reverse = True

        seg = self.copy()
        seg.points = seg.points[start:end+1]
        if reverse:
            seg.points = list(reversed(seg.points))

        return seg

    def copy(self):
        """ Creates a deep copy of this instance

        Returns:
            :obj:`Segment`
        """
        return deepcopy(self)

    def to_json(self):
        """ Converts segment to a JSON serializable format

        Returns:
            :obj:`dict`
        """
        points = [point.to_json() for point in self.points]
        return {
            'points': points,
            'transportationModes': self.transportation_modes,
            'locationFrom': self.location_from.to_json() if self.location_from != None else None,
            'locationTo': self.location_to.to_json() if self.location_to != None else None
        }

    @staticmethod
    def from_gpx(gpx_segment):
        """ Creates a segment from a GPX format.

        No preprocessing is done.

        Arguments:
            gpx_segment (:obj:`gpxpy.GPXTrackSegment`)
        Return:
            :obj:`Segment`
        """
        points = []
        for point in gpx_segment.points:
            points.append(Point.from_gpx(point))
        return Segment(points)

    @staticmethod
    def from_json(json):
        """ Creates a segment from a JSON file.

        No preprocessing is done.

        Arguments:
            json (:obj:`dict`): JSON representation. See to_json.
        Return:
            :obj:`Segment`
        """
        points = []
        for point in json['points']:
            points.append(Point.from_json(point))
        return Segment(points)
